/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
#include "jmdb.h"

/* history and input variables */
char	*history[MAX_HISTORY];
int		history_count = 0;
int		first_history = 0;
char	txt[256];
int 	txt_ptr;

/* alias info */
AliasInfo *aliases = NULL;

/* pipes and .jmdbinit file descriptors */
FILE	*wfp;
FILE	*rfp;
FILE	*jmdbinit;

extern FILE			*yyin;
extern char			debug_mode;
extern VariableInfo var;

/******************************************************************/
/*
	FUNCTION NAME:

		fgetline

	DESCRIPTION:

		read a line of input up to a `cr', EOF, or character limit

	FORMAL PARAMETERS:

		FILE *fp: file to read from
		char s[]: string to read into
		int lim:  maximum number of characters to read

	FUNCTION VALUE:

		returns number of characters read

*/
/******************************************************************/
int fgetline( fp, s, lim )
FILE *fp;
char s[];
int lim;
{
	int c, i;

	for( i = 0; i < lim - 1 && ( c = getc(fp) ) != EOF && c != '\n'; ++i )
		s[ i ] = c;
	if( c =='\n' )
		s[ i ] = '\0';

	if( c == EOF )
		return -1;
	else
		return i;
}

int getline( s, lim )
char s[];
int lim;
{
	int c, i;

	for( i = 0; i < lim - 1 && ( c = getchar() ) != EOF && c != '\n'; ++i )
		s[ i ] = c;
	if( c =='\n' )
		s[ i ] = '\0';
	return i;
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_error

	DESCRIPTION:

		error handler for user input to debugger

	FORMAL PARAMETERS:

		char *s: the string that caused the error

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_error(string)
char *string;
{
	int i;

	fprintf( stdout, "%s\n", txt );
	if( !strcmp( string, "\n" ) )
		for( i = 0; i < strlen( txt ); i++ )
			fprintf( stdout, " " );
	else
		for( i = 0; i < strlen( txt ) - strlen( string ) - 1; i++ )
			fprintf( stdout, " " );
	fprintf( stdout, STR_SYNTAX_ERROR );
}

/******************************************************************/
/*
	FUNCTION NAME:

		savetok

	DESCRIPTION:

		called from lex; stores all input

	FORMAL PARAMETERS:

		char *string: string to store

	FUNCTION VALUE:

		none

*/
/******************************************************************/
savetok( string )
char *string;
{
	txt[ txt_ptr ] = '\0';
	strcat( txt, string );
	strcat( txt, " " );
	txt_ptr = strlen( txt );
}

/******************************************************************/
/*
	FUNCTION NAME:

		store_command

	DESCRIPTION:

		stores the debugger command into a history buffer

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
store_command()
{
	static int index = 0;

	if( history_count >= MAX_HISTORY )
	{
		first_history = history_count - MAX_HISTORY + 1;
		free( history[ index ] );
	}

	history[ index ] = (char *)malloc( strlen( txt ) + 1 );
	if( history[ index ] == NULL )
	{
		fprintf( stdout, STR_MALLOC_ERR );
		abort();
	}
	strcpy( history[ index ], txt );
	history_count++;
	index++;
	if( index == MAX_HISTORY )
		index = 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_alias

	DESCRIPTION:

		handle the JMDB alias command

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_alias( num_toks, alias, command )
int num_toks;
char *alias;
char *command;
{
	AliasInfo *current_entry;

	/* Display all aliases */
	if( num_toks == 0 )
	{
		current_entry = aliases;
		while( current_entry != NULL )
		{
			fprintf( stdout,
				"%-8s %-32s\n", current_entry->alias, current_entry->command );
			current_entry = current_entry->next;
		}
	}
	/* Display alias */
	else if( num_toks == 1 )
	{
		current_entry = aliases;
		while( current_entry != NULL )
		{
			if( !strcmp( current_entry->alias, alias ) )
			{
				fprintf( stdout, "\"%s\"\n",current_entry->command );
				break;
			}
			current_entry = current_entry->next;
		}
		if( current_entry == NULL )
			fprintf( stdout, "\n" );
	}
	/*  Edit alias */
	else if( num_toks == 2 )
	{
		current_entry = aliases;
		while( current_entry != NULL )
		{
			if( !strcmp( current_entry->alias, alias ) )
			{
				free( current_entry->command );
				break;
			}
			current_entry = current_entry->next;
		}

		if( current_entry == NULL )
		{
			current_entry = (AliasInfo *)malloc( sizeof( AliasInfo ) );
			if( current_entry == NULL )
			{
				fprintf( stdout, STR_MALLOC_ERR );
				abort();
			}
			current_entry->alias = (char *)malloc( strlen( alias ) + 1 );
			if( current_entry->alias == NULL )
			{
				fprintf( stdout, STR_MALLOC_ERR );
				abort();
			}
			strcpy( current_entry->alias, alias );
			current_entry->next = aliases;
			aliases = current_entry;
		}

		current_entry->command = (char *)malloc( strlen( command ) + 1 );
		if( current_entry->command == NULL )
		{
			fprintf( stdout, STR_MALLOC_ERR );
			abort();
		}
		strcpy( current_entry->command, command );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_unalias

	DESCRIPTION:

		handle the JMDB unalias command

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_unalias( num_toks, alias )
int num_toks;
char *alias;
{
	AliasInfo *current_entry;
	AliasInfo *previous_entry;

	current_entry = aliases;
	previous_entry = NULL;
	while( current_entry != NULL )
	{
		/* Found and alias to delete */
		if( !strcmp( current_entry->alias, alias ) )
		{
			if( previous_entry == NULL )
				aliases = current_entry->next;
			else
				previous_entry->next = current_entry->next;
			free( current_entry->alias );
			free( current_entry->command );
			free( current_entry );
			return;
		}
		previous_entry = current_entry;
		current_entry = current_entry->next;
	}
	fprintf( stdout, STR_UNKNOWN_ALIAS, alias );
}

/******************************************************************/
/*
	FUNCTION NAME:

		is_alias

	DESCRIPTION:

		returns 1 if string is an alias
		returns 0 if string is not an alias

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
int
is_alias( command )
char *command;
{
	AliasInfo *current_entry;

	current_entry = aliases;
	while( current_entry != NULL )
	{
		if( !strcmp( current_entry->alias, command ) )
		{
			lex_unput( current_entry->command );
			return(1);
		}
		current_entry = current_entry->next;
	}
	return(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

		create_default_aliases

	DESCRIPTION:

		create JMDB default aliases

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
create_default_aliases()
{
	handle_alias( 2, "q", "quit" );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_history

	DESCRIPTION:

		display history of user commands

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_history()
{
	int i, index;

	index = first_history;
	while( index >= MAX_HISTORY )
		index = index - MAX_HISTORY;

	for( i = first_history; i < history_count; i++ )
	{
		fprintf( stdout, "%3d      %s\n", i + 1, history[ index ] );

		index++;
		if( index == MAX_HISTORY )
			index = 0;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_bang

	DESCRIPTION:

		handles the `bang' functionality
		allowing a user to re-execute a command

	FORMAL PARAMETERS:

		int number: the number of the command to execute

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_bang( number )
int number;
{
	int filedes[2];
	int index;

	if( number < 0 )
		index = history_count + number;
	else
		index = number - 1;

	if( ( index < first_history ) || ( index > ( history_count - 1 )  ) )
	{
		fprintf( stdout, STR_UNKNOWN_EVENT, index + 1 );
		return;
	}

	while( index >= MAX_HISTORY )
		index = index - MAX_HISTORY;

	txt_ptr = 0;

	pipe( filedes );
	wfp = fdopen( filedes[1], "w" );
	rfp = fdopen( filedes[0], "r" );

	fprintf( stdout, "%s\n", history[ index ] );
	fflush( stdout );
	fprintf( wfp, "%s\n", history[ index ] );
	fflush( wfp );

	yyin = rfp;
	yyparse();

	close_tmp_files();
}

/******************************************************************/
/*
	FUNCTION NAME:

		find_bang

	DESCRIPTION:

		searches for an event in the history buffer
		using a string key

	FORMAL PARAMETERS:

		char *string: the string to search for

	FUNCTION VALUE:

		the number of the event found

*/
/******************************************************************/
int find_bang( string )
char *string;
{
	int i, index;

	for( i = history_count - 1; i >= first_history; i-- )
	{
		index = i;
		while( index >= MAX_HISTORY )
			index = index - MAX_HISTORY;
		if( !strncmp( history[ index ], string, strlen( string ) ) )
			return( i + 1 );
	}
	return( 0 );
}

/******************************************************************/
/*
	FUNCTION NAME:

		read_jmdbinit

	DESCRIPTION:

		reads the .jmdbinit file when a user begins a jmdb session

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
read_jmdbinit()
{
	char line[64];
	int filedes[2];
	char filename[128];

	if( ( jmdbinit = fopen( ".jmdbinit", "r" ) ) == NULL )
	{
		sprintf( filename, "%s/%s", getenv("HOME"), ".jmdbinit" );

		if( ( jmdbinit = fopen( filename, "r" ) ) == NULL )
			return;
	}

	pipe( filedes );
	wfp = fdopen( filedes[1], "w" );
	rfp = fdopen( filedes[0], "r" );

	while( fgetline( jmdbinit, line, 64 ) != -1 )
	{
		if( line[0] != '#' )
		{
			txt_ptr = 0;
			fprintf( wfp, "%s\n", line );
			fflush( wfp );
	
			yyin = rfp;
			yyparse();
		}
	}
	close_tmp_files();
}

/******************************************************************/
/*
	FUNCTION NAME:

		close_tmp_files

	DESCRIPTION:

		closes all pipes and .jmdbinit

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
close_tmp_files()
{
	if( wfp != NULL )
	{
		fclose( wfp );
		wfp = NULL;
	}
	if( rfp != NULL )
	{
		fclose( rfp );
		rfp = NULL;
	}
	if( jmdbinit != NULL )
	{
		fclose( jmdbinit );
		jmdbinit = NULL;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		get_input

	DESCRIPTION:

		main input routine

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		returns 1: user wishes to exit jmdb
		returns 0: otherwise

*/
/******************************************************************/
int get_input()
{
	if( debug_mode )
	{
		if( var.prompt2 != NULL )
			fprintf( stdout, var.prompt2, history_count + 1 );
	}
	else
	{
		if( var.prompt1 != NULL )
			fprintf( stdout, var.prompt1, history_count + 1 );
	}
	fflush(stdout);
	txt_ptr = 0;
	yyin = stdin;
	return( yyparse() );
}
