/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
#include "jmdb.h"
#include "commands.h"
#include "../inc/time.h"

extern sigint_handler();
extern char Interrupt;

ClientInfo	client;
JmInfo		jm;
ProcessInfo	process;
VariableInfo var;
int rexec_sock;

int			num_procs;
char		debug_mode = 0;
char		login_name[32];
char		password[8];

/******************************************************************/
/*
	FUNCTION NAME:

		connected

	DESCRIPTION:

		returns `0' if `client' == NULL, otherwise returns 1

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		0: client == NULL
		1: client != NULL

*/
/******************************************************************/
int connected( handle )
CLIENT *handle;
{
	if( handle == NULL )
	{
		fprintf( stdout, STR_NOT_CONNECTED );
		return(0);
	}
	rpc_get_jm_status_1( NULL, handle );
	return(1);
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_rpc_failure

	DESCRIPTION:

		error handler called from routines in ../cps/librpcclnt.a

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_rpc_failure()
{
	fprintf( stdout, STR_FAILED_RPC_CALL );

	clnt_perror( client.cl, jm.hostname );

	handle_disconnect();

	siglongjmp( env, 1 );
}

/******************************************************************/
/*
	FUNCTION NAME:

		in_debug_mode

	DESCRIPTION:

		determine if in debug mode or not

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
char
in_debug_mode()
{
	if( !debug_mode )
	{
		fprintf( stdout, STR_WRONG_MODE );
		return( 0 );
	}
	else
		return( 1 );
}

/******************************************************************/
/*
	FUNCTION NAME:

		check_number_range

	DESCRIPTION:

		determines if number that user specified is within a certain range
		calls siglongjmp to reset the procedure stack

	FORMAL PARAMETERS:

		int number: number input
		int min: smallest number in range
		int max: largest number on range

	FUNCTION VALUE:

		none

*/
/******************************************************************/
check_number_range( number, min, max )
int number, min, max;
{
	if( ( number < min ) || ( number > max ) )
	{
		fprintf( stdout, STR_RANGE_ERR, number, min, max );
		siglongjmp( env, 1 );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		get_rexec_port

	DESCRIPTION:

		get a port number for use with rexec calls

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
int
get_rexec_port()
{
	static int port = 0;
	struct servent *serv;

	if( port ==  0 )
	{
		serv = getservbyname("exec","tcp");
		if( serv == NULL )
			port = 0;
		else
			port = serv->s_port;
	}
	return( port );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_commands

	DESCRIPTION:

		print list of JMDB commands

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_commands()
{
	int index = 0;

	while( strcmp( commands[ index ], "" ) )
	{
		fprintf( stdout, "%-16s", commands[ index++ ] );
		if( ( index % 4 ) == 0 )
			fprintf( stdout, "\n" );
	}
	if( ( index % 4 ) != 0 )
		fprintf( stdout, "\n" );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_help

	DESCRIPTION:

		print help file

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_help()
{
	char string[128];
	char *helpfile = "$CPSBIN/jmdb.help";

	sprintf( string, "%s %s", var.pager, helpfile );
	if( system( string ) < 0 )
		perror("help");
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_cps_mon

	DESCRIPTION:

		run cps_mon (acp_mon)

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_cps_mon(command)
char *command;
{
	char string[128];

	if( !connected( client.cl ) ) return;

	sprintf( string,"%s %s %s %d %d", command, "-a", jm.hostname,
	    client.prognum, client.version );
	if( system( string ) < 0 )
		perror("cps_mon");
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_kill

	DESCRIPTION:

		kill a user process or the entire job

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_kill( procnum )
int procnum;
{
	char string[128];

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	if( procnum == -1 )
	{
		rexec_sock = rexec_call( client.hostname, "" );
		if( rexec_sock < 0 )
		{
			fprintf( stdout, STR_KILL_ABORTED );
			return;
		}
		else
			close( rexec_sock );

		rpc_kill_jm_1( NULL, client.cl );
		client.cl = NULL;
	}
	else
	{
		check_number_range( procnum, 1, num_procs );
		sprintf( string, "%s %d", "kill -TRAP", process.pid[ procnum - 1 ] );

		rexec_sock = rexec_call( process.host[ procnum - 1 ], string );
		if( rexec_sock < 0 )
		{
			fprintf( stdout, STR_KILL_ABORTED );
			return;
		}
		else
			close( rexec_sock );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_process

	DESCRIPTION:

		list all processes in job

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_process()
{
	int i;
	char str[4][64];
	FILE *temp;
	char tempfile[128];
	char string[128];
	char *tmpstr;

	if( !connected( client.cl ) ) return;

	strcpy( tempfile, tempnam( "", "jmdb" ) );
	if( ( temp = fopen( tempfile, "w" ) ) == NULL )
	{
		fprintf( stdout, STR_FOPEN_ERR, tempfile, strerror(errno) );
		return;
	}

	fprintf( temp, "%-12s %-16s %-6s %-32s\n", STR_CLASS_PROC, STR_HOSTNAME, STR_PID, STR_PROGRAM );
	for( i = 0; i < num_procs; i++ )
	{
		sprintf( str[0], "[%d,%d]", process.class[ i ], i + 1 );
		sprintf( str[1], "%s", process.host[ i ] );
		if( *rpc_get_process_state_1( &i, client.cl ) == PSTATE_DEAD)
			sprintf( str[2], "%s", "Dead" );
		else
			sprintf( str[2], "%d", process.pid[ i ] );
		sprintf( str[3], "%s", process.name[ i ]);
		fprintf( temp, "%-12s %-16s %-6s %s\n", str[0], str[1],
		    str[2], str[3] );
	}
	fclose( temp );
	sprintf( string, "%s %s", var.pager, tempfile );
	if( system( string ) < 0 )
		perror("process");
	unlink( tempfile );
}

handle_ping()
{
	int index;
	char string[128];

	for( index = 0; index < num_procs; index++ )
	{
		sprintf( string, "%s %s %s",
			"/usr/etc/ping", "-c 1", process.host[ index ] );
		if( system( string ) < 0 )
			perror("ping");
	}
}

handle_rexec()
{
	char *hostnames[256];
	int num_hosts = 0;
	int index, host_index;
	char host_found;

	if( !connected( client.cl ) ) return;

	for( index = 0; index < num_procs; index++ )
	{
		host_found = 0;
		for( host_index = 0; host_index < num_hosts; host_index++ )
			if( !strcmp( hostnames[ host_index ], process.host[ index ] ) )
				host_found = 1;

		if( host_found )
			continue;
		else
		{
			hostnames[ num_hosts ] =
				(char *)malloc( strlen( process.host[ index ] ) );
			strcpy( hostnames[ num_hosts++ ], process.host[ index ] );
		}

		fprintf( stdout, STR_CONTACTING_HOST, process.host[ index ] );

		rexec_sock = rexec_call( process.host[ index ], "" );
		if( rexec_sock < 0 )
			perror("rexec");
		else
		{
			fprintf( stdout, "O.K.\n" );
			close( rexec_sock );
		}
	}
	for( host_index = 0; host_index < num_hosts; host_index++ )
		free( hostnames[ host_index ] );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_dbx

	DESCRIPTION:

		run dbx on a user process or the JM

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_dbx( procnum, obj_name )
int procnum;
char *obj_name;
{
	char string[128];
	char objfile[128];
	char corefile[128];
	char obj_hostname[256];
	char obj_program[256];
	char work_directory[256];
	int obj_pid;
	int error = FALSE;
	FILE *fp;

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	if( procnum == 0 )
	{
		strcpy( obj_hostname, jm.hostname );
		strcpy( work_directory, jm.directory );
		obj_pid = jm.pid;
	}
	else
	{
		check_number_range( procnum, 1, num_procs );
		strcpy( obj_hostname, process.host[ procnum - 1 ] );

		if( obj_name != NULL )
			strcpy( obj_program, obj_name );

		strcpy( work_directory, process.directory[ procnum - 1 ] );
		obj_pid = process.pid[ procnum - 1 ];
	}

	if( var.dbxcopy || strcmp( client.hostname, obj_hostname ) )
	{
		fprintf( stdout, STR_CONTACTING_HOST, obj_hostname );

		/* contact remote host */
		sprintf( string, "%s %d", "kill -USR2", obj_pid );

		rexec_sock = rexec_call( obj_hostname, string );
		if( rexec_sock < 0 )
		{
			fprintf( stdout, STR_DBX_ABORTED );
			return;
		}
		else
			close( rexec_sock );

		/* copy core file from process working directory */
		if( var.coredir == NULL )
		{
			sprintf( string,"%s %s:%s/core %s","rcp", obj_hostname,
			    work_directory, var.dbgdir );
			fprintf( stdout, STR_COPYING_CORE );
			fprintf( stdout, "%s\n", string );
			if( system( string ) < 0 )
			{
				error = TRUE;
				perror("rcp");
			}
			sprintf( corefile, "%s/%s", var.dbgdir, "core" );
		}
		else
		{
			sprintf( corefile, "%s/%s", var.coredir, "core" );
		}

		/* make sure core file exists and is readable */
		if( ( fp = fopen( corefile, "r" ) ) == NULL )
		{
			error = TRUE;
			fprintf( stdout, STR_FOPEN_ERR, corefile, strerror(errno) );
		}
		else
		{
#ifdef SGI
			if( obj_name == NULL )
			{
				fseek( fp, 16, SEEK_SET );
				fgets( obj_program, 81, fp );
			}
#endif
			fclose( fp );
		}
		if( error )
		{
			fprintf( stdout, STR_DBX_ABORTED );
			return;
		}

		/* copy executable after reading name from core file */
		if( var.objdir == NULL )
		{
			if( procnum == 0 )
				sprintf( string,"%s %s:%s/%s %s","rcp", obj_hostname,
				    getenv("CPSBIN"), obj_program, var.dbgdir );
			else
				if( obj_name == NULL )
					sprintf( string,"%s %s:%s/%s %s","rcp", obj_hostname,
					    work_directory, obj_program, var.dbgdir );
				else
					sprintf( string,"%s %s:%s %s","rcp", obj_hostname,
					    obj_program, var.dbgdir );
			fprintf( stdout, STR_COPYING_EXECUTABLE );
			fprintf( stdout, "%s\n", string );
			if( system( string ) < 0 )
			{
				error = TRUE;
				perror("rcp");
			}
			sprintf( objfile, "%s/%s", var.dbgdir,
				obj_program );
		}
		else
		{
			sprintf( objfile, "%s/%s", var.objdir,
				obj_program );
		}

		/* make sure object file exists and is readable */
		if( ( fp = fopen( objfile, "r" ) ) == NULL )
		{
			error = TRUE;
			fprintf( stdout, STR_FOPEN_ERR, objfile, strerror(errno) );
		}
		else
			fclose( fp );
		if( error )
		{
			if( var.coredir == NULL )
			{
				fprintf( stdout, STR_REMOVING_CORE );
				unlink( corefile );
			}
			fprintf( stdout, STR_DBX_ABORTED );
			return;
		}

		/* run dbx using core file */
		sprintf( string,"%s %s %s",var.dbx, objfile, corefile );
		if( system( string ) < 0 )
		{
			perror("dbx");
			fprintf( stdout, STR_DBX_ABORTED );
		}

		if( var.coredir == NULL )
		{
			fprintf( stdout, STR_REMOVING_CORE );
			unlink( corefile );
		}
	}
	/* run dbx on running process */
	else
	{
		if( obj_name != NULL )
			fprintf( stdout, "Ignoring %s\n", obj_name );
		fprintf( stdout, STR_CONNECTING_TO_PROCESS );
		sprintf( string,"%s %d", var.dbxrun, obj_pid );
		if( system( string ) < 0 )
		{
			perror("dbx");
			fprintf( stdout, STR_DBX_ABORTED );
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_jminfo_all

	DESCRIPTION:

		run jminfo on a list of hostnames

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_jminfo_all()
{
	char *hostname;
	char *token;

	if( var.hosts == NULL )
	{
		fprintf( stdout, STR_HOSTS_NOT_SET );
		return;
	}

	token = (char *)malloc( strlen( var.hosts ) + 1 );
	strcpy( token, var.hosts );

	fprintf( stdout, "%-4s %-16s %-8s %-32s\n",
		STR_ID, STR_HOSTNAME, STR_USER, STR_JOB );

	hostname = (char *)strtok( token, " " );
	handle_jminfo( hostname, 1 );

	while( ( hostname = (char *)strtok( NULL, " " ) ) != NULL )
		handle_jminfo( hostname, 1 );

	free( token );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_jminfo

	DESCRIPTION:

		query hostname for CPS 2.9 or later jobs

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_jminfo( hostname, type )
char *hostname;
int type;
{
	char string[128];
	CLIENT *handle;
	u_long prognum;
	u_long version = 1;
	int job;
	char str[5][32];
	static char res;
	static struct timeval TIMEOUT = { 5, 0 	};
	struct sockaddr_in addr;
	struct hostent *hp;
	int sock;
	char firsttime = 1;

	hp = gethostbyname( hostname );
	if( hp == NULL )
	{
		fprintf( stdout, "%s:%s\n", hostname, strerror( errno ) );
		return;
	}
	bcopy( hp->h_addr, &addr.sin_addr, hp->h_length );
	addr.sin_family = AF_INET;

	if( type == 0 )
		fprintf( stdout, STR_CONTACTING_HOST, hostname );
	for( prognum = MIN_PROGNUM; prognum <= MAX_PROGNUM; prognum++ )
	{
		Interrupt = 0;
		/* handle = clnt_create( hostname, prognum, version, "tcp" ); */
		addr.sin_port = 0;
		sock = RPC_ANYSOCK;
		handle = clnttcp_create( &addr, prognum, version, &sock, 0, 0 );
		if( Interrupt == 1 )
			continue;
		if( handle != NULL )
		{
			handle->cl_auth = authunix_create_default();
			bzero((char *)&res, sizeof(res));
			if( clnt_call( handle, RPC_GET_JM_STATUS, xdr_void, 0,
			    xdr_char, &res, TIMEOUT ) == RPC_SUCCESS )
			{
				job = prognum - MIN_PROGNUM;
				if( firsttime )
				{
					if( type == 0 )
						fprintf( stdout, "%-4s %-16s %-8s %-32s\n",
							STR_ID, STR_HOSTNAME, STR_USER, STR_JOB );
					firsttime = 0;
				}
				sprintf( str[0], "%d", job );
				sprintf( str[1], "%s", hostname );
				sprintf( str[2], "%d", prognum );
				sprintf( str[3], "%s", *rpc_get_job_name_1( NULL, handle ) );
				sprintf( str[4], "%s", *rpc_get_jm_username_1( NULL, handle ) );
				fprintf( stdout, "%-4s %-16s %-8s %-32s\n",
				    str[0], str[1], str[4], str[3] );
			}
			auth_destroy( handle->cl_auth );
			clnt_destroy( handle );
		}
	}
	if( firsttime && ( type == 0 ) )
		fprintf( stdout, STR_NO_JMS_FOUND, hostname );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_job

	DESCRIPTION:

		print miscellaneous job info

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_job()
{
	int flag = 2;

	if( !connected( client.cl ) ) return;

	fprintf( stdout, "%-12s %s\n", STR_HOSTNAME,
	    jm.hostname );
	fprintf( stdout, "%-12s %d\n", STR_ID,
	    client.prognum - MIN_PROGNUM );
	fprintf( stdout, "%-12s %s\n", STR_USER,
	    jm.username );
	fprintf( stdout, "%-12s %s\n", STR_JOB,
	    jm.jobname );
	fprintf( stdout, "%-12s %s", STR_START,
	    ctime( &jm.starttime ) );
	fprintf( stdout, "%-12s %s\n", STR_VERSION,
	    jm.version );

	fprintf( stdout, "%-12s", STR_STATE );
	if( *rpc_get_jm_status_1( NULL, client.cl ) )
		fprintf( stdout, " %s\n", STR_STOPPED );
	else
		fprintf( stdout, " %s\n", STR_RUNNING );

	fprintf( stdout, "%-12s", STR_UPSHOT );
	if( *rpc_upshot_switch_1( &flag, client.cl ) )
		fprintf( stdout, STR_ON );
	else
		fprintf( stdout, STR_OFF );

	fprintf( stdout, "%-12s", STR_HOOKS );
	if( *rpc_hook_switch_1( &flag, client.cl ) )
		fprintf( stdout, STR_ON );
	else
		fprintf( stdout, STR_OFF );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_queues

	DESCRIPTION:

		print state of all declared queues

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_queues()
{
	int q;
	int_array_res *result;
	int_array     plist;
	int found = 0;
	int ndq;

	if( !connected( client.cl ) ) return;

	for( q = 0; q < MAX_QUEUES; q++ )
		if( *rpc_get_q_declared_1( &q, client.cl ) )
		{
			if( !found )
			{
				fprintf( stdout, "%-5s %-4s %-9s\n",
					STR_QUEUE, "#PDQ", STR_PROCESS );
				fprintf( stdout, "%-5s %-4s %-7s\n",
					"-----", "----", "-------");
			}

			fprintf( stdout, "   %d ", q );

			ndq = *rpc_get_ndq_set_1( &q, client.cl );
			fprintf( stdout, "  %2d   ", ndq );

			result = rpc_get_process_queue_1( &q, client.cl );

			for( plist = result->int_array_res_u.plist;
			    plist != NULL; plist = plist->next )
				fprintf( stdout, " %d", plist->lpn );

			fprintf( stdout, "\n" );
			found++;
		}

	if( !found )
		fprintf( stdout, STR_NO_DECLARED_QUEUES );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_syncs

	DESCRIPTION:

		print state of all sync points

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_syncs()
{
	int sync;
	int_array_res *result;
	int_array plist;
	int found = 0;
	int index;
	char at_sync[MAX_PROCESSES];
	char need_sync[MAX_PROCESSES];

	if( !connected( client.cl ) ) return;

	for( sync = 0; sync < MAX_SYNCS; sync++ )
	{
		result = rpc_get_need_sync_1( &sync, client.cl );
		if( result->errno == 0 )
		{
			if( !found )
			{
				fprintf( stdout, "%s %s\n", STR_SYNC, STR_PROCESS );
				fprintf( stdout, "---- -------\n" );
			}

			for( index = 0; index < MAX_PROCESSES; index++ )
			{
				at_sync[ index ] = 0;
				need_sync[ index ] = 0;
			}

			for( plist = result->int_array_res_u.plist;
			    plist != NULL; plist = plist->next )
				need_sync[ plist->lpn ] = 1;

			result = rpc_get_at_sync_1( &sync, client.cl );
			if( result->errno == 0 )
			{
				for( plist = result->int_array_res_u.plist;
				    plist != NULL; plist = plist->next )
					at_sync[ plist->lpn ] = 1;

				fprintf( stdout, "%4d", sync );
				for( index = 0; index < MAX_PROCESSES; index++ )
				{
					if( need_sync[ index ] )
					{
						fprintf( stdout, " %d", index );
						if( at_sync[ index ] )
							fprintf( stdout, "x" );
					}
				}

				fprintf( stdout, "\n" );
			}
			found++;
		}
	}
	if( !found )
		fprintf( stdout, STR_NO_SYNC_POINTS );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_state

	DESCRIPTION:

		print state of all proceses

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_state( flag )
int flag;
{
	int i;
	char str[2][64];
	int state, attrib;
	FILE *temp;
	char tempfile[128];
	char string[128];

	if( !connected( client.cl ) ) return;

	if( flag == -1 )
	{
		if( *rpc_get_jm_status_1( NULL, client.cl ) )
			fprintf( stdout, "%-12s %s\n", "[JM]", STR_STOPPED );
		else
			fprintf( stdout, "%-12s %s\n", "[JM]", STR_RUNNING );
		return;
	}

	strcpy( tempfile, tempnam( "", "jmdb" ) );
	if( ( temp = fopen( tempfile, "w" ) ) == NULL )
	{
		fprintf( stdout, STR_FOPEN_ERR, tempfile, strerror(errno) );
		return;
	}

	fprintf( temp, "%s %s\n", STR_CLASS_PROC, STR_STATE);

	for( i = 0; i < num_procs; i++ )
	{
		sprintf( str[0], "[%d,%d]", process.class[ i ], i + 1 );
		state = *rpc_get_process_state_1( &i, client.cl );
		attrib = *rpc_get_process_attributes_1( &i, client.cl );
		switch( state )
		{
		case PSTATE_RUNNING :
			sprintf(str[1],"Running");
			break;
		case PSTATE_DEAD :
			sprintf(str[1],"Dead");
			break;
		case PSTATE_FINISHED :
			sprintf(str[1],"Finished");
			break;
		case PSTATE_WAIT_SYNC :
			sprintf(str[1],"Waiting on sync number %d",attrib);
			break;
		case PSTATE_WAIT_QUEUE_FULL :
			sprintf(str[1],"Waiting for queue %d to become full",attrib);
			break;
		case PSTATE_WAIT_QUEUE_EMPTY :
			sprintf(str[1],"Waiting for queue %d to become empty",attrib);					
			break;
		case PSTATE_WAIT_DEQUEUE :
			sprintf(str[1],"Waiting to dequeue from queue %d",attrib);
			break;
		case PSTATE_WAIT_CALL :
			sprintf(str[1],"Waiting for remote sub %d to return",attrib);
			break;
		case PSTATE_WAIT_GET :
			sprintf(str[1],"Waiting for cps_get to complete, block=%d",attrib);
			break;
		case PSTATE_QUEUED :
			sprintf(str[1],"Queued on %d",attrib);
			break;
		case PSTATE_WAIT_SERVICE :
			sprintf(str[1],"Waiting to service call");
			break;
		case PSTATE_WAIT_XPL :
			sprintf(str[1],"Waiting for explicit message %x",attrib);
			break;
		case PSTATE_WAIT_SEND :
			sprintf(str[1],"Waiting for cps_send to complete, block=%d",attrib);
			break;
		case PSTATE_WAIT_MOUNT :
			sprintf(str[1],"Waiting for mount reply");
			break;
		} /* end of switch */
		fprintf( temp, "%-12s %s\n", str[0], str[1] );
	} /* end of for loop */
	fclose( temp );
	sprintf( string, "%s %s", var.pager, tempfile );
	if( system( string ) < 0 )
		perror("state");
	unlink( tempfile );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_upshot_switch

	DESCRIPTION:

		turn upshot logging on or off

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_upshot_switch( flag )
int flag;
{
	int status;
	int u_flag;

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	u_flag = flag;
	status = *rpc_upshot_switch_1( &u_flag, client.cl );
	if( flag == 0 )
	{
		if( status == 0 )
			fprintf( stdout, STR_UPSHOT_LOG_OFF, jm.hostname, jm.jobname );
		else if( status == 1 )
			fprintf( stdout, STR_UPSHOT_LOG_NOT_ON );
	}
	else if( flag == 1 )
	{
		if( status == 0 )
			fprintf( stdout, STR_UPSHOT_LOG_ON );
		else if( status == 1 )
			fprintf( stdout, STR_UPSHOT_LOG_NOT_OFF );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_upshot

	DESCRIPTION:

		run upshot

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_upshot_show(command)
char *command;
{
	char string[128];

	sprintf( string, "%s %s %s.up", command, "-l", jm.jobname );
	if( system( string ) < 0 )
		perror("upshot");
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_hooks_switch

	DESCRIPTION:

		turn hooks on or off

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_hooks_switch( flag )
int flag;
{
	int status;
	int u_flag;

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	u_flag = flag;
	status = *rpc_hook_switch_1( &u_flag, client.cl );
	if( flag == 0 )
	{
		if( status == 0 )
			fprintf( stdout, STR_PHOOKS_OFF );
		else if( status == 1 )
			fprintf( stdout, STR_PHOOKS_NOT_ON );
	}
	else if( flag == 1 )
	{
		if( status == 0 )
			fprintf( stdout, STR_PHOOKS_ON );
		else if( status == 1 )
			fprintf( stdout, STR_PHOOKS_NOT_OFF );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		set_string

	DESCRIPTION:

		set a JMDB character string variable

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
set_string( variable, value )
char **variable;
char *value;
{
	if( *variable != NULL )
		free( *variable );

	if( strcmp( value, "(null)" ) )
	{
		*variable = (char *)malloc( strlen( value ) + 1 );
		if( *variable == NULL )
		{
			fprintf( stdout, STR_MALLOC_ERR );
			abort();
		}
		strcpy( *variable, value );
	}
	else
		*variable = NULL;
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_set

	DESCRIPTION:

		set a JMDB variable

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_set( type, integer, string )
int type;
int integer;
char *string;
{
	static char *on_off[2] = { 
		"off", "on" 	};
	char *tmpstring;

	switch( type )
	{
	case VAR_LIST:
		fprintf( stdout, "\nSTRINGS\n" );
		fprintf( stdout, "%-16s %s\n", "coredir", var.coredir );
		fprintf( stdout, "%-16s %s\n", "dbgdir", var.dbgdir );
		fprintf( stdout, "%-16s %s\n", "dbx", var.dbx );
		fprintf( stdout, "%-16s %s\n", "dbxrun", var.dbxrun );
		fprintf( stdout, "%-16s %s\n", "hosts", var.hosts );
		fprintf( stdout, "%-16s %s\n", "objdir", var.objdir );
		/*
		fprintf( stdout, "%-16s %s\n", "output", var.output );
		*/
		fprintf( stdout, "%-16s %s\n", "pager", var.pager );
		fprintf( stdout, "%-16s %s\n", "prompt1", var.prompt1 );
		fprintf( stdout, "%-16s %s\n", "prompt2", var.prompt2 );
		fprintf( stdout, "\nFLAGS\n" );
		fprintf( stdout, "%-16s %s\n", "dbxcopy", on_off[ var.dbxcopy ] );
		/*
		fprintf( stdout, "%-16s %s\n", "display", on_off[ var.display ] );
		*/
		fprintf( stdout, "%-16s %s\n", "queues", on_off[ var.queues ] );
		fprintf( stdout, "%-16s %s\n", "syncs", on_off[ var.syncs ] );
		break;
	case VAR_SYNCS:
		var.syncs = integer;
		break;
	case VAR_QUEUES:
		var.queues = integer;
		break;
	case VAR_DISPLAY:
		var.display = integer;
		break;
	case VAR_DBXCOPY:
		var.dbxcopy = integer;
		break;
	case VAR_COREDIR:
		set_string( &var.coredir, string );
		break;
	case VAR_OBJDIR:
		set_string( &var.objdir, string );
		break;
	case VAR_DBGDIR:
		set_string( &var.dbgdir, string );
		break;
	case VAR_OUTPUT:
		set_string( &var.output, string );
		break;
	case VAR_PAGER:
		set_string( &var.pager, string );
		break;
	case VAR_DBX:
		set_string( &var.dbx, string );
		break;
	case VAR_DBXRUN:
		set_string( &var.dbxrun, string );
		break;
	case VAR_PROMPT1:
		set_string( &var.prompt1, string );
		break;
	case VAR_PROMPT2:
		set_string( &var.prompt2, string );
		break;
	case VAR_HOSTS:
		if( strcmp( string, "(null)" ) )
		{
			if( var.hosts == NULL )
			{
				set_string( &var.hosts, string );
			}
			else
			{
				tmpstring = (char *)malloc( strlen( var.hosts ) +
					strlen( string ) + 1 );
				sprintf( tmpstring, "%s %s", var.hosts, string );
				set_string( &var.hosts, tmpstring );
				free( tmpstring );
			}
		}
		else
			set_string( &var.hosts, string );
		break;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_jml

	DESCRIPTION:

		display jml (not implemented)

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_jml( flag )
int flag;
{
	char log_file[128];
	char jml_file[256];
	char string[256];
	char line[128];
	FILE *rexec_stream;

	if( !connected( client.cl ) ) return;

	strcpy( log_file, *rpc_get_log_file_1( NULL, client.cl ) );
	if( strncmp( log_file, "/", 1 ) )
		sprintf( jml_file, "%s/%s", jm.directory, log_file );
	else
		sprintf( jml_file, "%s", log_file );

	switch( flag )
	{
	case 0:
		sprintf( string, "cat %s", jml_file );
		break;
	case 1:
		sprintf( string, "tail -f %s", jml_file );
		break;
	}

	rexec_sock = rexec_call( jm.hostname, string );
	if( rexec_sock < 0 )
	{
		perror("rexec");
		fprintf( stdout, STR_FAILED_TO_CONNECT );
		return;
	}

	rexec_stream = fdopen( rexec_sock, "r" );
	while( fgetline( rexec_stream, line, 128 ) != -1 )
		fprintf( stdout, "%s\n", line );

	close( rexec_sock );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_debug

	DESCRIPTION:

		enter debug mode

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_debug()
{
	if( debug_mode )
		return;

	if( !connected( client.cl ) )
		return;

	debug_mode = 1;
	handle_import();
}

int
handle_exit()
{
	if( debug_mode )
	{
		debug_mode = 0;
		return( 0 );
	}
	else
		return( handle_quit( 0 ) );
}

/******************************************************************/
/*
	FUNCTION NAME:

		dump_info

	DESCRIPTION:

		auto display of syncs and queues

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
dump_info()
{
	if( var.queues )
		handle_queues();
	if( var.syncs )
		handle_syncs();
	if( var.display )
		qmon_update( 1 );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_quit

	DESCRIPTION:

		quit from JMDB

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
int
handle_quit( force )
int force;
{
	char quit_prompt = 0;
	char reply[4];

	if( client.cl == NULL )
		return( 1 );

	if( !force )
	{
		if( breakpoint_exists() )
		{
			fprintf( stdout, STR_BREAKPOINTS_EXIST );
			quit_prompt = 1;
		}
		else if( client.cl != NULL && *rpc_get_jm_status_1( NULL, client.cl ) )
		{
			fprintf( stdout, STR_JM_IS_STOPPED );
			quit_prompt = 1;
		}
	}

	if( quit_prompt )
	{
		fgetline( stdin, reply, 4 );
		if( !strncmp( reply, "y", 1 ) )
			return( 1 );
		else
			return( 0 );
	}
	else
		return( 1 );
}

/******************************************************************/
/*
	FUNCTION NAME:

		connect_to_server

	DESCRIPTION:

		connect to a JMDB server

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
connect_to_server( job, host, prog, vers )
char *job;
char *host;
u_long prog;
u_long vers;
{
	CLIENT *cl;
	int status;
	int i;
	char *tmpstr;
	struct hostent *hp;

	if( vers == 0 )
		read_prog_file( job, host, &prog, &vers );

	if( prog < MIN_PROGNUM )
		prog = prog + MIN_PROGNUM;

	if( ( hp = gethostbyname( host ) ) == NULL )
	{
		fprintf( stdout, STR_GETHOSTBYNAME_ERR, host, strerror(errno) );
		fprintf( stdout, STR_FAILED_TO_CONNECT );
		return;
	}
	strcpy( host, hp->h_name );

	if( ( jm.hostname != NULL ) &&
	    ( !strcmp( host, jm.hostname ) ) &&
	    ( prog == client.prognum ) )
	{
		rpc_get_jm_status_1( NULL, client.cl );
		fprintf( stdout, STR_ALREADY_CONNECTED );
		return;
	}

	login_prompt( host );

	rexec_sock = rexec_call( host, "" );
	if( rexec_sock < 0 )
	{
		perror("rexec");
		fprintf( stdout, STR_FAILED_TO_CONNECT );
		return;
	}
	close( rexec_sock );

	create_rpc_client( job, host, &prog, &vers, &cl );
	if( cl == NULL )
	{
		fprintf( stdout, STR_FAILED_TO_CONNECT );
		return;
	}

	destroy_rpc_client( client.cl );

	init_break_points();
	init_declared_queues();

	client.prognum = prog;
	client.version = vers;
	client.cl = cl;

	strcpy( jm.hostname, host );

	strcpy( jm.version, *rpc_get_jm_version_1( NULL, client.cl ) );

	strcpy( jm.username, *rpc_get_jm_username_1( NULL, client.cl ) );

	strcpy( jm.directory, *rpc_get_jm_directory_1( NULL, client.cl ) );

	jm.pid = *rpc_get_jm_pid_1( NULL, client.cl );

	jm.starttime = *rpc_get_jm_starttime_1( NULL, client.cl );

	num_procs = *rpc_get_num_proc_1( NULL, client.cl );

	for( i = 0; i < num_procs; i++ )
	{
		process.pid[ i ] = *rpc_get_process_id_1( &i, client.cl );
		process.class[ i ] = *rpc_get_process_class_1( &i, client.cl );

		tmpstr = *rpc_get_process_host_1( &i, client.cl );
		process.host[ i ] = (char *)malloc( strlen( tmpstr ) + 1 );
		if( process.host[ i ] == NULL )
		{
			fprintf( stdout, STR_MALLOC_ERR );
			abort();
		}
		strcpy( process.host[ i ], tmpstr );

		tmpstr = *rpc_get_process_name_1( &i, client.cl );
		process.name[ i ] = (char *)malloc( strlen( tmpstr ) + 1 );
		if( process.name[ i ] == NULL )
		{
			fprintf( stdout, STR_MALLOC_ERR );
			abort();
		}
		strcpy( process.name[ i ], tmpstr );

		tmpstr = *rpc_get_process_dir_1( &i, client.cl );
		process.directory[ i ] = (char *)malloc( strlen( tmpstr ) + 1 );
		if( process.directory[ i ] == NULL )
		{
			fprintf( stdout, STR_MALLOC_ERR );
			abort();
		}
		strcpy( process.directory[ i ], tmpstr );
	}

	fprintf( stdout, STR_CONNECTED_TO );
	handle_job();
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_disconnect

	DESCRIPTION:

		disconnect from job

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_disconnect()
{
	destroy_rpc_client( client.cl );
	client.prognum = 0;
	client.cl = NULL;
	debug_mode = 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

		jmdb_init

	DESCRIPTION:

		initialize JMDB

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
jmdb_init()
{
	struct hostent *hp;

	if( gethostname( client.hostname, 64 ) == -1 )
	{
		fprintf( stdout, STR_GETHOSTNAME_ERR, client.hostname, strerror(errno) );
		exit(1);
	}
	if( ( hp = gethostbyname( client.hostname ) ) == NULL )
	{
		fprintf( stdout, STR_GETHOSTBYNAME_ERR, client.hostname, strerror(errno) );
		exit(1);
	}
	strcpy( client.hostname, hp->h_name );

	handle_set( VAR_DBGDIR, 0, "/usr/tmp" );
	handle_set( VAR_PROMPT1, 0, "(jmdb:%d) " );
	handle_set( VAR_PROMPT2, 0, "[jmdb:%d] " );
#ifdef SGI
	handle_set( VAR_PAGER, 0, "/usr/bsd/more" );
	handle_set( VAR_DBX, 0, "/usr/bin/dbx" );
	handle_set( VAR_DBXRUN, 0, "/usr/bin/dbx -p" );
#endif
#ifdef HPUX
	handle_set( VAR_PAGER, 0, "/usr/bin/more" );
	handle_set( VAR_DBX, 0, "/usr/bin/dbx" );
	handle_set( VAR_DBXRUN, 0, "/usr/bin/dbx -p" );
#endif
#ifdef AIX
	handle_set( VAR_PAGER, 0, "/bin/more" );
	handle_set( VAR_DBX, 0, "/bin/dbx" );
	handle_set( VAR_DBXRUN, 0, "/bin/dbx -a" );
#endif
#ifdef ULTRIX
	handle_set( VAR_PAGER, 0, "/usr/ucb/more" );
	handle_set( VAR_DBX, 0, "/usr/ucb/dbx" );
#endif

	init_allowable_function_set();

	read_jmdbinit();
}

/******************************************************************/
/*
	FUNCTION NAME:

		jmdb_exit

	DESCRIPTION:

		terminate JMDB

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
jmdb_exit()
{
	destroy_rpc_client( client.cl );
	exit(0);
}

int
rexec_call( hostname, command )
char *hostname;
char *command;
{
	int sock;
	char *host, **ahost;

	host = hostname;
	ahost = &host;

	sock = rexec( ahost, get_rexec_port(), login_name, password, command, 0 );

	/* AIX BUG PATCH - rexec resets the signal hander - very bad! */
	if( sig_establish_handler( SIGINT, sigint_handler ) )
	{
		fprintf( stdout, STR_SIGNAL_HANDLER_ERR, "SIGINT" );
		exit( 1 );
	}

	return( sock );
}

login_prompt( hostname )
char *hostname;
{
	char *pass;
	char *user_name;
	char pass_string[64];
	char name_string[64];

	user_name = (char *)cuserid(0);

	sprintf( name_string, "Name (%s:%s): ", hostname, user_name );
	printf( "%s", name_string );
	getline( login_name, 32 );
	if( strlen( login_name ) == 0 )
		strcpy( login_name, user_name );

	if( strcmp( login_name, user_name ) )
		sprintf( pass_string, "Password (%s:%s): ", hostname, login_name );
	else
		sprintf( pass_string, "Password (%s:%s): ", hostname, user_name );
	pass = (char *)getpass( pass_string );
	strcpy( password, pass );
	/*
	if( strlen( password ) == 0 )
		return( 1 );
	*/
}
