/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.	    
	   All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or	     
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/******************************************************************************/
/*										  */
/*										  */
/*				CPS Monitor					  */
/*										  */
/*										  */
/******************************************************************************/

/*****************/
/* Include Files */
/*****************/

#include <curses.h>
#include "../inc/perr.h"
#include "../inc/types.h"
#include "../inc/param.h"
#include "../inc/mon.h"
#include "../inc/xfunc.h"
#include <stdio.h>
#include <sys/socket.h>
#include <rpc/rpc.h>
#include <signal.h>
#include <errno.h>
#include "../rpc/jm_rpc.h"

#ifdef HPUX
#include <unistd.h>
#endif

/**********/
/* Macros */
/**********/

#define FACILITY "MON"
#define MIN_PROGNUM 0x40000000

#define USAGE "\
Usage: cps_mon [-dufsqt] [-c <class>] [-p <proc>] job_name\n\
         or\n\
       cps_mon [-dufsqt] [-c <class>] [-p <proc>] -h|-a hostname prognum\n"

#define XUSAGE "\
Usage: xcps_mon [options]\n\
\n\
OPTION           X-RESOURCE   DESCRIPTION\n\
-a <hostname>    *hostname    same as '-h' option.\n\
-c <n>           *class       monitor processes in class <n>.\n\
-d                            monitor delta calls.\n\
-f                            monitor calls.\n\
-fn <fontname>   *fontList    Font to use in monitor.\n\
-font <fontname> *fontList    same as '-fn' option.\n\
-g                            monitor delta call graph.\n\
-h <hostname>    *hostname    Connect to job on machine <hostname>.\n\
-i <n>           *id          Connect to job ID <n>.\n\
-id <n>          *id          same as '-i' option.\n\
-interval <n>    *interval    set polling time to <n> seconds.\n\
-m <type>        *monitor     <type> is \"deltas\",\"user\",\"calls\",\"states\",\n\
                              \"queues\",\"rates\",\"graph\"\n\
-monitor <type>  *monitor     same as '-m' option.\n\
-p <proc>        *process     monitor process number <proc>.\n\
-q                            monitor queues.\n\
-s                            monitor process states.\n\
-t                            monitor transfer rates.\n\
-u                            monitor user status.\n\
<filename>                    name of 'jdf' file without suffix.\n"


extern char xcps_mon;
char monitor_name[16];

/**********************************/
/* External Non-Integer Functions */
/**********************************/

FILE *fopen();
CLIENT *cl;
char *strchr();

/********************/
/* Global Variables */
/********************/

char xcpsmon = FALSE;
char rpc_failure = 0;
char check_for_access_rights = FALSE;

extern int optind;
extern char *optarg;

int mon_class = -1;	/* All */
int mon_proc = -1;	/* All */
int mon_type = MON_CALLS; 
char rpcvers = FALSE;

char job_name[128];
char job_hostname[128];

struct cps_job_stats *mon_shm;
struct cps_job_stats mon_loc;

int shmid;

WINDOW *pad;
int pad_rows,pad_cols;
int pad_x = 0;
int pad_y = 0;

int init_done = 0;

extern int suppress_errors;

int job_done = 0;

int calls[ 6 ] =
{
	INC_CALL, INC_SEND, INC_GET, INC_QUEUE, INC_DEQUEUE, INC_CALLS_SERV
};

/****************************/
/* Display Field Structures */
/****************************/

struct {
	int x[MAX_USER_STATUS],y[MAX_USER_STATUS];
} status_field[MAX_PROCESSES];

struct {
	int x[12],y[12];
} call_cnt[MAX_PROCESSES];

struct {
	int l;
	int x,y;
} pstate[MAX_PROCESSES];

struct {
	int l;
	int x,y;
} qloc[MAX_QUEUES];
int next_q_row;

int num_dead_procs = 0;
int dead_proc[MAX_PROCESSES];


/********************/
/* Simple Functions */
/********************/

char is_dead( proc )
int proc;
{
	int index;
	for( index = 0; index < num_dead_procs; index++ )
		if( dead_proc[ index ] == proc )
			return( TRUE );
	return( FALSE );
}

void UpdateGraph( i, proc, type, draw )
int i, proc, type;
char draw;
{
	static int offset = 0;
	int_pair pair;
	int x;

	if( ( type == -1 ) || ( type == 999 ) )
	{
		if( ( type == 999 ) && ( offset == 0 ) )
		{
			if( rpcvers )
				x = *rpc_get_process_state_1( &i, cl );
			else
				x = mon_shm->p_state[i].state;
			if( x == PSTATE_DEAD )
			{
				XcpsmonUpdateDeadProcess( 0, proc );
				dead_proc[ num_dead_procs++ ] = i;
			}
		}
		offset = 0;
		if( draw )
			XcpsmonUpdateProcessGraph( 0, proc, 0, type, 0 );
		return;
	}

	pair.x = i;
	pair.y = type;
	if( rpcvers )
		x = *rpc_get_call_count_1( &pair, cl );
	else
		x = mon_shm->call_cnt[i][pair.y];

	if( mon_loc.call_cnt[i][type] != x )
	{
		if( draw )
			XcpsmonUpdateProcessGraph( 0, proc,
				x - mon_loc.call_cnt[i][type], type, offset );
		offset += ( x - mon_loc.call_cnt[i][type] );
		mon_loc.call_cnt[i][type] = x;
	}
}

display_error(str)
char *str;
{
	if( !xcpsmon )
	{
		if (init_done)
		{
		    werase(pad);
		    mon_refresh();
		    delwin(pad);
		    endwin();
		}
	}
	printf("%s\n",str);
	exit(0);
}


handle_rpc_failure()
{
	if( check_for_access_rights )
	{
		printf("%s: Permission Denied\n",monitor_name);
		exit(0);
	}
	rpc_failure = 1;
	if( !xcpsmon )
	{
		echo_str(pad_rows-2,34,"Lost Connection with RPC server");
		pause();
	}
	else
	{
		XcpsmonUpdateFooter( "Lost Connection with RPC server");
		XcpsmonDeactivateMenus();
	}
}

mon_refresh() {

	if( rpcvers && rpc_failure )
		return;

	if( !xcpsmon )
#ifdef ULTRIX
		refresh();
#else
		prefresh(pad,pad_y,pad_x,0,0,24,80);
#endif
	else
		XcpsmonForceUpdates();
}

void handle_ctrlc(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	mon_exit();
}

void handle_alarm(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	if( rpcvers && rpc_failure )
		return;

	check_for_updates();
	if( !job_done )
	{
		if( !xcpsmon )
			alarm(5);
		else
			XcpsmonTimeout();
	}
}

mon_exit()
{
	if( !xcpsmon )
	{
		if (init_done) {
			werase(pad);
			mon_refresh();
			delwin(pad);
			endwin();
		}
	}
	exit(0);
}

int mon_this_proc(i)
int i;
{
	if ((mon_proc != -1) && (mon_proc != (i+1))) return(0);
	if ((mon_class != -1) && (mon_class != mon_loc.class[i])) return(0);
	return(-1);
}

echo_str(y,x,str)
int y,x;
char *str;
{

	if( !xcpsmon )
	{
#if defined(ULTRIX) || defined(AIX) || defined(HPUX) || defined(SUNOS)

		mvwaddstr(pad,y,x,str);
		mon_refresh();

#else

		wmove(pad,y,x);
		while (*str != '\0') {
			pechochar(pad,*str);
			str++;
		}

#endif
	}

}

/******************************************************************/
/*
	FUNCTION NAME:

	 force_updates

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

force_updates()
{

	int i,j,n;
	int_pair pair;
	char temp[80];
	int x;
	int attrib;
	int index;
	int proc = 0;

	next_q_row = 4;

	if( rpcvers && rpc_failure )
		return;

	switch (mon_type)
	{
		case MON_USER_STATUS :

/* Check user status values */

			for (i=0; i<mon_loc.num_proc; i++)
			{
				if (!mon_this_proc(i)) continue;

				pair.x = i;
				for (j=0; j<MAX_USER_STATUS; j++)
				{
					pair.y = j;
					if( rpcvers )
						x = *rpc_get_user_status_1( &pair, cl );
					else
						x = mon_shm->user_status[i][j];

					if( !xcpsmon )
					{
						sprintf(temp,"%10d",x);
						echo_str(status_field[i].y[j],
							status_field[i].x[j],temp);
					}
					else
					{
						sprintf(temp,"%d",x);
						XcpsmonUpdateProcessLine( j, proc, temp );
					}
					mon_loc.user_status[i][j] = x;
				}
				proc++;
			}
			break;
	case MON_CALLS :

/* Check call counts */

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;

			pair.x = i;
			for( index = 0; index < 6; index++ )
			{
				pair.y = calls[ index ];
				if( rpcvers )
					x = *rpc_get_call_count_1( &pair, cl );
				else
					x = mon_shm->call_cnt[i][pair.y];

				if( !xcpsmon )
				{
					sprintf(temp,"%10d",x);
					echo_str(call_cnt[i].y[pair.y],
						call_cnt[i].x[pair.y],temp);
				}
				else
				{
					sprintf(temp,"%d",x);
					XcpsmonUpdateProcessLine( index, proc, temp );
				}
				mon_loc.call_cnt[i][pair.y] = x;
			}
			proc++;
		}

		break;
	case MON_DELTA_CALLS :

/* Check delta call counts */

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;

			pair.x = i;
			for( index = 0; index < 6; index++ )
			{
				pair.y = calls[ index ];
				if( rpcvers )
					x = *rpc_get_call_count_1( &pair, cl );
				else
					x = mon_shm->call_cnt[i][pair.y];

				if( !xcpsmon )
				{
					sprintf(temp,"%10d",0);
					echo_str(call_cnt[i].y[pair.y],
						call_cnt[i].x[pair.y],temp);
				}
				else
				{
					sprintf(temp,"%d",0);
					XcpsmonUpdateProcessLine( index, proc, temp );
				}
				mon_loc.call_cnt[i][pair.y] = x;
			}
			proc++;
		}

		break;

	case MON_DELTA_GRAPH:

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;
			UpdateGraph( i, proc, -1, 0 );
			UpdateGraph( i, proc, INC_CALL, 0 );
			UpdateGraph( i, proc, INC_SEND, 0 );
			UpdateGraph( i, proc, INC_GET, 0 );
			UpdateGraph( i, proc, INC_QUEUE, 0 );
			UpdateGraph( i, proc, INC_DEQUEUE, 0 );
			UpdateGraph( i, proc, INC_CALLS_SERV, 0 );
			UpdateGraph( i, proc++, 999, 0 );
		}

		break;

	case MON_STATES :

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;

			if( rpcvers )
			{
				x = *rpc_get_process_state_1( &i, cl );
				attrib = *rpc_get_process_attributes_1( &i, cl );
			}
			else
			{
				x = mon_shm->p_state[i].state;
				attrib = mon_shm->p_state[i].attrib;
			}

			switch(x)
			{
				case PSTATE_RUNNING :
					sprintf(temp,"Running");
					break;
				case PSTATE_DEAD :
					sprintf(temp,"Dead");
					break;
				case PSTATE_FINISHED :
					sprintf(temp,"Finished");
					break;
				case PSTATE_WAIT_SYNC :
					sprintf(temp,"Waiting on sync number %d",attrib);
					break;
				case PSTATE_WAIT_QUEUE_FULL :
					sprintf(temp,"Waiting for queue %d to become full",attrib);
					break;
				case PSTATE_WAIT_QUEUE_EMPTY :
					sprintf(temp,"Waiting for queue %d to become empty",attrib);
					break;
				case PSTATE_WAIT_DEQUEUE :
					sprintf(temp,"Waiting to dequeue from queue %d",attrib);
					break;
				case PSTATE_WAIT_CALL :
					sprintf(temp,"Waiting for remote sub %d to return",attrib);
					break;
				case PSTATE_WAIT_GET :
					sprintf(temp,"Waiting for cps_get to complete, block=%d",attrib);
					break;
				case PSTATE_QUEUED :
					sprintf(temp,"Queued on %d",attrib);
					break;
				case PSTATE_WAIT_SERVICE :
					sprintf(temp,"Waiting to service call");
					break;
				case PSTATE_WAIT_XPL :
					sprintf(temp,"Waiting for explicit message %x",attrib);
					break;
				case PSTATE_WAIT_SEND :
					sprintf(temp,"Waiting for cps_send to complete, block=%d",attrib);
					break;
				case PSTATE_WAIT_MOUNT :
					sprintf(temp,"Waiting for mount reply");
					break;
			}
			n = strlen(temp);
			for (j=n; j<pstate[i].l; j++) temp[j] = ' ';
			if (n < pstate[i].l) temp[pstate[i].l] = '\0';
			pstate[i].l = n;
			if( !xcpsmon )
				echo_str(pstate[i].y,pstate[i].x,temp);
			else
				XcpsmonUpdateProcessLine( 0, proc++, temp );
		}

		break;

	case MON_TRATES :

		for (i=0; i<4; i++)
		{
			if( rpcvers )
			{
				mon_loc.nsends[i] = *rpc_get_nsends_1( &i, cl );
				mon_loc.sndcur[i] = *rpc_get_sndcur_1( &i, cl );
				mon_loc.sndavg[i] = *rpc_get_sndavg_1( &i, cl );
				mon_loc.ngets[i] = *rpc_get_ngets_1( &i, cl );
				mon_loc.getcur[i] = *rpc_get_getcur_1( &i, cl );
				mon_loc.getavg[i] = *rpc_get_getavg_1( &i, cl );
			}
			else
			{
				mon_loc.nsends[i] = mon_shm->nsends[i];
				mon_loc.sndcur[i] = mon_shm->sndcur[i];
				mon_loc.sndavg[i] = mon_shm->sndavg[i];
				mon_loc.ngets[i] = mon_shm->ngets[i];
				mon_loc.getcur[i] = mon_shm->getcur[i];
				mon_loc.getavg[i] = mon_shm->getavg[i];
			}

			if( !xcpsmon )
			{
				sprintf(temp,"%7d",mon_loc.nsends[i]);
				echo_str(5+i*2,20,temp);

				sprintf(temp,"%7d",mon_loc.sndcur[i]);
				echo_str(5+i*2,30,temp);

				sprintf(temp,"%7d",mon_loc.sndavg[i]);
				echo_str(5+i*2,40,temp);

				sprintf(temp,"%7d",mon_loc.ngets[i]);
				echo_str(5+i*2,50,temp);

				sprintf(temp,"%7d",mon_loc.getcur[i]);
				echo_str(5+i*2,60,temp);

				sprintf(temp,"%7d",mon_loc.getavg[i]);
				echo_str(5+i*2,70,temp);
			}
			else
			{
				sprintf(temp,"%d",mon_loc.nsends[i]);
				XcpsmonUpdateProcessLine( 0, i, temp );

				sprintf(temp,"%d",mon_loc.sndcur[i]);
				XcpsmonUpdateProcessLine( 1, i, temp );

				sprintf(temp,"%d",mon_loc.sndavg[i]);
				XcpsmonUpdateProcessLine( 2, i, temp );

				sprintf(temp,"%d",mon_loc.ngets[i]);
				XcpsmonUpdateProcessLine( 3, i, temp );

				sprintf(temp,"%d",mon_loc.getcur[i]);
				XcpsmonUpdateProcessLine( 4, i, temp );

				sprintf(temp,"%d",mon_loc.getavg[i]);
				XcpsmonUpdateProcessLine( 5, i, temp );
			}
		}
		break;

	case MON_QUEUES :

/* Check for newly declared queues */

		for (i=0; i<MAX_QUEUES; i++)
		{
			if( rpcvers )
				x = *rpc_get_q_declared_1( &i, cl );
			else
				x = mon_shm->q_declared[i];
			if( x )
			{
				mon_loc.q_declared[i] = -1;
				qloc[i].x = 12;
				qloc[i].y = next_q_row;
				qloc[i].l = 0;
				sprintf(temp,"[%d]",i);
				if( !xcpsmon )
					echo_str(next_q_row,0,temp);
				else
					XcpsmonProcessLine( temp, 0 );
				next_q_row++;
			}
		}
		break;
	}

}

/******************************************************************/
/*
	FUNCTION NAME:

	 check_for_updates

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

check_for_updates()
{

	int i,j,n;
	int delta;
	char temp[80];
	int x;
	int attrib;
	int_pair pair;
	int index;
	int *z;
	int proc = 0;


	if( rpcvers )
	{
		x = *rpc_get_job_done_1( NULL, cl );
		if( rpc_failure )
			return;
		if( x )
		{
			if( !xcpsmon )
				echo_str(pad_rows-2,34,"Job Finished");
			else
			{
				XcpsmonUpdateFooter( "Job Finished" );
				XcpsmonDeactivateMenus();
			}
			job_done = -1;
		}
	}

	switch (mon_type)
	{

	case MON_USER_STATUS :

/* Check user status values */

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;

			pair.x = i;
			for (j=0; j<MAX_USER_STATUS; j++)
			{
				pair.y = j;
				if( rpcvers )
					x = *rpc_get_user_status_1( &pair, cl );
				else
					x = mon_shm->user_status[i][j];

				if (mon_loc.user_status[i][j] != x)
				{
					if( !xcpsmon )
					{
						sprintf(temp,"%10d",x);
						echo_str(status_field[i].y[j],
							status_field[i].x[j],temp);
					}
					else
					{
						sprintf(temp,"%d",x);
						XcpsmonUpdateProcessLine( j, proc, temp );
					}
					mon_loc.user_status[i][j] = x;
				}
			}
			proc++;
		}
		break;

	case MON_CALLS :

/* Check call counts */

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;

			pair.x = i;
			for( index = 0; index < 6; index++ )
			{
				pair.y = calls[ index ];
				if( rpcvers )
					x = *rpc_get_call_count_1( &pair, cl );
				else
					x = mon_shm->call_cnt[i][pair.y];

				if (mon_loc.call_cnt[i][pair.y] != x)
				{
					if( !xcpsmon )
					{
						sprintf(temp,"%10d",x);
						echo_str(call_cnt[i].y[pair.y],
							call_cnt[i].x[pair.y],temp);
					}
					else
					{
						sprintf(temp,"%d",x);
						XcpsmonUpdateProcessLine( index, proc, temp );
					}
					mon_loc.call_cnt[i][pair.y] = x;
				}
			}
			proc++;
		}

		break;

	case MON_DELTA_CALLS :

/* Check call counts */

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;

			pair.x = i;
			for( index = 0; index < 6; index++ )
			{
				pair.y = calls[ index ];
				if( rpcvers )
					x = *rpc_get_call_count_1( &pair, cl );
				else
					x = mon_shm->call_cnt[i][pair.y];

				if( !xcpsmon )
				{
					sprintf(temp,"%10d",x - mon_loc.call_cnt[i][pair.y]);
					echo_str(call_cnt[i].y[pair.y],
						call_cnt[i].x[pair.y],temp);
				}
				else
				{
					sprintf(temp,"%d",x - mon_loc.call_cnt[i][pair.y]);
					XcpsmonUpdateProcessLine( index, proc, temp );
				}
				mon_loc.call_cnt[i][pair.y] = x;
			}
			proc++;
		}

		break;

	case MON_DELTA_GRAPH:

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;
			if( !is_dead( i ) )
			{
				UpdateGraph( i, proc, -1, 1 );
				UpdateGraph( i, proc, INC_CALL, 1 );
				UpdateGraph( i, proc, INC_SEND, 1 );
				UpdateGraph( i, proc, INC_GET, 1 );
				UpdateGraph( i, proc, INC_QUEUE, 1 );
				UpdateGraph( i, proc, INC_DEQUEUE, 1 );
				UpdateGraph( i, proc, INC_CALLS_SERV, 1 );
				UpdateGraph( i, proc, 999, 1 );
			}
			proc++;
		}

		break;

	case MON_STATES :

		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;

			if( rpcvers )
			{
				x = *rpc_get_process_state_1( &i, cl );
				attrib = *rpc_get_process_attributes_1( &i, cl );
			}
			else
			{
				x = mon_shm->p_state[i].state;
				attrib = mon_shm->p_state[i].attrib;
			}
			if ((mon_loc.p_state[i].state != x) || 
				(mon_loc.p_state[i].attrib != attrib))
			{
				mon_loc.p_state[i].state = x;
				mon_loc.p_state[i].attrib = attrib;

				switch(x) {
				case PSTATE_RUNNING :
					sprintf(temp,"Running");
					break;
				case PSTATE_DEAD :
					sprintf(temp,"Dead");
					break;
				case PSTATE_FINISHED :
					sprintf(temp,"Finished");
					break;
				case PSTATE_WAIT_SYNC :
					sprintf(temp,"Waiting on sync number %d",attrib);
					break;
				case PSTATE_WAIT_QUEUE_FULL :
					sprintf(temp,"Waiting for queue %d to become full",attrib);
					break;
				case PSTATE_WAIT_QUEUE_EMPTY :
					sprintf(temp,"Waiting for queue %d to become empty",attrib);
					break;
				case PSTATE_WAIT_DEQUEUE :
					sprintf(temp,"Waiting to dequeue from queue %d",attrib);
					break;
				case PSTATE_WAIT_CALL :
					sprintf(temp,"Waiting for remote sub %d to return",attrib);
					break;
				case PSTATE_WAIT_GET :
					sprintf(temp,"Waiting for cps_get to complete, block=%d",attrib);
					break;
				case PSTATE_QUEUED :
					sprintf(temp,"Queued on %d",attrib);
					break;
				case PSTATE_WAIT_SERVICE :
					sprintf(temp,"Waiting to service call");
					break;
				case PSTATE_WAIT_XPL :
					sprintf(temp,"Waiting for explicit message %x",attrib);
					break;
				case PSTATE_WAIT_SEND :
					sprintf(temp,"Waiting for cps_send to complete, block=%d",attrib);
					break;
				case PSTATE_WAIT_MOUNT :
					sprintf(temp,"Waiting for mount reply");
					break;
				}
				n = strlen(temp);
				for (j=n; j<pstate[i].l; j++) temp[j] = ' ';
				if (n < pstate[i].l) temp[pstate[i].l] = '\0';
				pstate[i].l = n;
				if( !xcpsmon )
					echo_str(pstate[i].y,pstate[i].x,temp);
				else
					XcpsmonUpdateProcessLine( 0, proc++, temp );
			}

		}
		break;

	case MON_TRATES :

		for (i=0; i<4; i++)
		{
			if( rpcvers )
			{
				mon_shm->nsends[i] = *rpc_get_nsends_1( &i, cl );
				mon_shm->sndcur[i] = *rpc_get_sndcur_1( &i, cl );
				mon_shm->sndavg[i] = *rpc_get_sndavg_1( &i, cl );
				mon_shm->ngets[i] = *rpc_get_ngets_1( &i, cl );
				mon_shm->getcur[i] = *rpc_get_getcur_1( &i, cl );
				mon_shm->getavg[i] = *rpc_get_getavg_1( &i, cl );
			}

			if (mon_loc.nsends[i] != mon_shm->nsends[i])
			{
				mon_loc.nsends[i] = mon_shm->nsends[i];
				if( !xcpsmon )
				{
					sprintf(temp,"%7d",mon_loc.nsends[i]);
					echo_str(5+i*2,20,temp);
				}
				else
				{
					sprintf(temp,"%d",mon_loc.nsends[i]);
					XcpsmonUpdateProcessLine( 0, i, temp );
				}
		  	}
			if (mon_loc.sndcur[i] != mon_shm->sndcur[i])
			{
				mon_loc.sndcur[i] = mon_shm->sndcur[i];
				if( !xcpsmon )
				{
					sprintf(temp,"%7d",mon_loc.sndcur[i]);
					echo_str(5+i*2,30,temp);
				}
				else
				{
					sprintf(temp,"%d",mon_loc.sndcur[i]);
					XcpsmonUpdateProcessLine( 1, i, temp );
				}
			}
			if (mon_loc.sndavg[i] != mon_shm->sndavg[i]) {
				mon_loc.sndavg[i] = mon_shm->sndavg[i];
				if( !xcpsmon )
				{
					sprintf(temp,"%7d",mon_loc.sndavg[i]);
					echo_str(5+i*2,40,temp);
				}
				else
				{
					sprintf(temp,"%d",mon_loc.sndavg[i]);
					XcpsmonUpdateProcessLine( 2, i, temp );
				}
			}

			if (mon_loc.ngets[i] != mon_shm->ngets[i])
			{
				mon_loc.ngets[i] = mon_shm->ngets[i];
				if( !xcpsmon )
				{
					sprintf(temp,"%7d",mon_loc.ngets[i]);
					echo_str(5+i*2,50,temp);
				}
				else
				{
					sprintf(temp,"%d",mon_loc.ngets[i]);
					XcpsmonUpdateProcessLine( 3, i, temp );
				}
			}
			if (mon_loc.getcur[i] != mon_shm->getcur[i])
			{
				mon_loc.getcur[i] = mon_shm->getcur[i];
				if( !xcpsmon )
				{
					sprintf(temp,"%7d",mon_loc.getcur[i]);
					echo_str(5+i*2,60,temp);
				}
				else
				{
					sprintf(temp,"%d",mon_loc.getcur[i]);
					XcpsmonUpdateProcessLine( 4, i, temp );
				}
			}
			if (mon_loc.getavg[i] != mon_shm->getavg[i])
			{
				mon_loc.getavg[i] = mon_shm->getavg[i];
				if( !xcpsmon )
				{
					sprintf(temp,"%7d",mon_loc.getavg[i]);
					echo_str(5+i*2,70,temp);
				}
				else
				{
					sprintf(temp,"%d",mon_loc.getavg[i]);
					XcpsmonUpdateProcessLine( 5, i, temp );
				}
			}
		}
		break;

	case MON_QUEUES :

/* Check for newly declared queues */

		for (i=0; i<MAX_QUEUES; i++)
		{
			if( rpcvers )
				x = *rpc_get_q_declared_1( &i, cl );
			else
				x = mon_shm->q_declared[i] && (!mon_loc.q_declared[i]);
			if( x && (!mon_loc.q_declared[i]) )
			{
				mon_loc.q_declared[i] = -1;
				qloc[i].x = 12;
				qloc[i].y = next_q_row;
				qloc[i].l = 0;
				sprintf(temp,"[%d]",i);
				if( !xcpsmon )
					echo_str(next_q_row,0,temp);
				else
					XcpsmonUpdateProcessLine( 0, next_q_row - 4, temp );
				next_q_row++;
			}
		}

/* Check for changes in declared queues */

		for (i=0; i<MAX_QUEUES; i++)
		{
			if (mon_loc.q_declared[i])
			{
				if( rpcvers )
					x = *rpc_get_q_set_1( &i, cl );
				else
					x = mon_shm->qset[i];
				if( x != mon_loc.qset[i] )
				{
					mon_loc.qset[i] = x;
					temp[0] = '\0';
					n = 0;
					z = set_query(&mon_loc.qset[i],8*sizeof(int));
				    while (*z != 0)
					{
						x = *z;
						z++;
						if (*z == 0)
							sprintf(&temp[n],"%d",x);
						else
							sprintf(&temp[n],"%d,",x);
					 	n = strlen(temp);
					}
					n = strlen(temp);
					for (j=n; j<qloc[i].l; j++) temp[j] = ' ';
					if (n < qloc[i].l) temp[qloc[i].l] = '\0';
					qloc[i].l = n;
					if( !xcpsmon )
						echo_str(qloc[i].y,qloc[i].x,temp);
					else
						XcpsmonUpdateProcessLine( 0, qloc[i].y - 4, temp );
				}
			}
		}
		break;
	}

	if( xcpsmon )
		XcpsmonForceUpdates();

/* Check to see if the job has finished */

	if( rpcvers )
		x = *rpc_get_job_done_1( NULL, cl );
	else
		x = mon_shm->job_done;
	if( x )
	{
		if( !xcpsmon )
			echo_str(pad_rows-2,34,"Job Finished");
		else
		{
			XcpsmonUpdateFooter( "Job Finished" );
			XcpsmonDeactivateMenus();
		}
		job_done = -1;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	 mon_init

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

mon_init() {

	char title[128],*mtype,mclass[4],mproc[4],temp[128];

	int i,j,r,c;

	int x,y;

	if( rpcvers && rpc_failure )
		return;

/* Create the pad */

	if( !xcpsmon )
	{
		initscr();
		pad_rows = mon_loc.num_proc+8;
		pad_rows = pad_rows > 24 ? pad_rows : 24;
		pad_cols = 80;
#ifdef ULTRIX
		pad = stdscr;
#else
		pad = newpad(pad_rows,pad_cols);
#endif
		noecho();
	}

/* Build the banner */

	if (mon_class == -1)
		strcpy(mclass,"all");
	else
		sprintf(mclass,"%d",mon_class);

	if (mon_proc == -1)
		strcpy(mproc,"all");
	else
		sprintf(mproc,"%d",mon_proc);

	switch (mon_type)
	{
		case MON_USER_STATUS :
			mtype = "User Status";
			break;
		case MON_CALLS :
			mtype = "Calls";
			break;
		case MON_STATES :
			mtype = "Process States";
			break;
		case MON_QUEUES :
			mtype = "Queues";
			break;
		case MON_TRATES :
			mtype = "Transfer Rates";
			break;
		case MON_DELTA_CALLS :
			mtype = "Delta Calls";
			break;
		case MON_DELTA_GRAPH :
			mtype = "Activity";
			break;
		default :
			mtype = "????";
	}

	if( !xcpsmon )
	{
		for (i=0; i<79; i++) title[i] = ' ';
		sprintf(title,"CPS Monitor %s",mtype);
		title[strlen(title)] = ' ';
		sprintf(temp,"Job: %s  Class: %s  Proc: %s",job_name,mclass,mproc);
		i = 79 - strlen(temp);
		strcpy(&title[i],temp);
		mvwaddstr(pad,0,0,title);
	}
	else
	{
		if( rpcvers )
			sprintf( title, "%s:%s - %s", job_hostname, job_name, mtype );
		else
			sprintf( title, "%s - %s", job_name, mtype );

		XcpsmonTitle( title );
	}

	switch (mon_type)
	{

	case MON_USER_STATUS :

		if( !xcpsmon )
			mvwaddstr(pad,2,0,"[Class,Proc]");
		else
		{
			XcpsmonWorkForm( MAX_USER_STATUS, 0 );
			XcpsmonFooter( "Running..." );
			XcpsmonColumnHeader( "[Class,Proc]", 1 );
		}
		c = 20;
		for (i=0; i<MAX_USER_STATUS; i++)
		{
			sprintf(temp,"(%d)",i+1);
			if( !xcpsmon )
				mvwaddstr(pad,2,c,temp);
			else
			{
				char temp2[100];
				sprintf( temp2, "%s", temp );
				XcpsmonColumnHeader( temp2, 1 );
				if( i == ( MAX_USER_STATUS - 1 ) )
					XcpsmonColumnHeader( NULL, 1 );
			}
			c += 10;
		}

		y = 4;
		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;
			sprintf(temp,"[%d,%d]",mon_loc.class[i],i+1);
			if( !xcpsmon )
				mvwaddstr(pad,y,0,temp);
			else
				XcpsmonProcessLine( temp, 1 );
			x = 13;
			for (j=0; j<MAX_USER_STATUS; j++)
			{
				status_field[i].y[j] = y;
				status_field[i].x[j] = x;
				x += 10;
			}
			y++;
		}

		break;

	case MON_CALLS :

		if( !xcpsmon )
		{
			mvwaddstr(pad,2,0,"[Class,Proc]");
			mvwaddstr(pad,2,12+(10-strlen("call")),"call");
			mvwaddstr(pad,2,22+(10-strlen("get")),"send");
			mvwaddstr(pad,2,32+(10-strlen("send")),"get");
			mvwaddstr(pad,2,42+(10-strlen("queue")),"queue");
			mvwaddstr(pad,2,52+(10-strlen("dequeue")),"dequeue");
			mvwaddstr(pad,2,64,"service_calls");
		}
		else
		{
			XcpsmonWorkForm( 6, 0 );
			XcpsmonFooter( "Running..." );
			XcpsmonColumnHeader( "[Class,Proc]", 1 );
			XcpsmonColumnHeader( "call", 1 );
			XcpsmonColumnHeader( "send", 1 );
			XcpsmonColumnHeader( "get", 1 );
			XcpsmonColumnHeader( "queue", 1 );
			XcpsmonColumnHeader( "dequeue", 1 );
			XcpsmonColumnHeader( "service_calls", 1 );
			XcpsmonColumnHeader( NULL, 1 );
		}

		y = 4;
		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;
			sprintf(temp,"[%d,%d]",mon_loc.class[i],i+1);
			if( !xcpsmon )
				mvwaddstr(pad,y,0,temp);
			else
				XcpsmonProcessLine( temp, 1 );
			call_cnt[i].x[INC_CALL] = 12;
			call_cnt[i].x[INC_SEND] = 22;
			call_cnt[i].x[INC_GET] = 32;
			call_cnt[i].x[INC_QUEUE] = 42;
			call_cnt[i].x[INC_DEQUEUE] = 52;
			call_cnt[i].x[INC_CALLS_SERV] = 64;
			call_cnt[i].y[INC_CALL] = y;
			call_cnt[i].y[INC_SEND] = y;
			call_cnt[i].y[INC_GET] = y;
			call_cnt[i].y[INC_QUEUE] = y;
			call_cnt[i].y[INC_DEQUEUE] = y;
			call_cnt[i].y[INC_CALLS_SERV] = y;
			y++;
		}

		break;

	case MON_DELTA_CALLS :

		if( !xcpsmon )
		{
			mvwaddstr(pad,2,0,"[Class,Proc]");
			mvwaddstr(pad,2,12+(10-strlen("call")),"call");
			mvwaddstr(pad,2,22+(10-strlen("get")),"send");
			mvwaddstr(pad,2,32+(10-strlen("send")),"get");
			mvwaddstr(pad,2,42+(10-strlen("queue")),"queue");
			mvwaddstr(pad,2,52+(10-strlen("dequeue")),"dequeue");
			mvwaddstr(pad,2,64,"service_calls");
		}
		else
		{
			XcpsmonWorkForm( 6, 0 );
			XcpsmonFooter( "Running..." );
			XcpsmonColumnHeader( "[Class,Proc]", 1 );
			XcpsmonColumnHeader( "call", 1 );
			XcpsmonColumnHeader( "send", 1 );
			XcpsmonColumnHeader( "get", 1 );
			XcpsmonColumnHeader( "queue", 1 );
			XcpsmonColumnHeader( "dequeue", 1 );
			XcpsmonColumnHeader( "service_calls", 1 );
			XcpsmonColumnHeader( NULL, 1 );
		}

		y = 4;
		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;
			sprintf(temp,"[%d,%d]",mon_loc.class[i],i+1);
			if( !xcpsmon )
				mvwaddstr(pad,y,0,temp);
			else
				XcpsmonProcessLine( temp, 1 );
			call_cnt[i].x[INC_CALL] = 12;
			call_cnt[i].x[INC_SEND] = 22;
			call_cnt[i].x[INC_GET] = 32;
			call_cnt[i].x[INC_QUEUE] = 42;
			call_cnt[i].x[INC_DEQUEUE] = 52;
			call_cnt[i].x[INC_CALLS_SERV] = 64;
			call_cnt[i].y[INC_CALL] = y;
			call_cnt[i].y[INC_SEND] = y;
			call_cnt[i].y[INC_GET] = y;
			call_cnt[i].y[INC_QUEUE] = y;
			call_cnt[i].y[INC_DEQUEUE] = y;
			call_cnt[i].y[INC_CALLS_SERV] = y;
			y++;
		}

		break;

	case MON_DELTA_GRAPH :

		XcpsmonWorkForm( 1, 1 );
		XcpsmonFooter( "Running..." );
		XcpsmonGraphColumnHeader( "[Class,Proc]" );
		XcpsmonGraphColumnHeader( "Activity" );
		XcpsmonGraphColumnHeader( NULL );
		for( i = 0; i < mon_loc.num_proc; i++ )
		{
			if (!mon_this_proc(i)) continue;
			sprintf(temp,"[%d,%d]",mon_loc.class[i],i+1);
			XcpsmonProcessGraph( temp );
		}
		XcpsmonGraphLegendElement( INC_CALL );
		XcpsmonGraphLegendElement( INC_SEND );
		XcpsmonGraphLegendElement( INC_GET );
		XcpsmonGraphLegendElement( INC_QUEUE );
		XcpsmonGraphLegendElement( INC_DEQUEUE );
		XcpsmonGraphLegendElement( INC_CALLS_SERV );
		XcpsmonGraphLegendElement( 0 );

		break;

	case MON_STATES :

		if( !xcpsmon )
		{
			mvwaddstr(pad,2,0,"[Class,Proc]");
			mvwaddstr(pad,2,14,"Process State");
		}
		else
		{
			XcpsmonWorkForm( 1, 0 );
			XcpsmonFooter( "Running..." );
			XcpsmonColumnHeader( "[Class,Proc]", 0 );
			XcpsmonColumnHeader( "Process State", 0 );
			/*
			XcpsmonColumnHeader( "                                      ", 0 );
			*/
			XcpsmonColumnHeader( NULL, 0 );
		}
		y = 4;
		for (i=0; i<mon_loc.num_proc; i++)
		{
			if (!mon_this_proc(i)) continue;
			sprintf(temp,"[%d,%d]",mon_loc.class[i],i+1);
			if( !xcpsmon )
				mvwaddstr(pad,y,0,temp);
			else
				XcpsmonProcessLine( temp, 0 );
			pstate[i].l = 0;
			pstate[i].x = 14;
			pstate[i].y = y;
			y++;
		}
		break;

	case MON_QUEUES :

		if( !xcpsmon )
		{
			mvwaddstr(pad,2,0,"[Queue]");
			mvwaddstr(pad,2,12,"Processes Queued");
		}
		else
		{
			XcpsmonWorkForm( 1, 0 );
			XcpsmonFooter( "Running..." );
			XcpsmonColumnHeader( "[Queue]", 0 );
			XcpsmonColumnHeader( "Processes Queued", 0 );
			XcpsmonColumnHeader( NULL, 0 );
		}
		break;

	case MON_TRATES :

		if( !xcpsmon )
		{
			mvwaddstr(pad,2,5,"Transfer Size");
			mvwaddstr(pad,2,20,"#sends");
			mvwaddstr(pad,2,30,"cps_send");
			mvwaddstr(pad,3,30,"cur. bps");
			mvwaddstr(pad,2,40,"cps_send");
			mvwaddstr(pad,3,40,"avg. bps");
			mvwaddstr(pad,2,50,"#gets");
			mvwaddstr(pad,2,60,"cps_get");
			mvwaddstr(pad,3,60,"cur. bps");
			mvwaddstr(pad,2,70,"cps_get");
			mvwaddstr(pad,3,70,"avg. bps");
			mvwaddstr(pad,5,5,"< 2K");
			mvwaddstr(pad,7,5,"< 8K");
			mvwaddstr(pad,9,5,"< 32K");
			mvwaddstr(pad,11,5,">= 32K");
		}
		else
		{
			XcpsmonWorkForm( 6, 0 );
			XcpsmonColumnHeader( "Transfer Size", 1 );
			XcpsmonColumnHeader( "#sends", 1 );
			XcpsmonColumnHeader( "cps_send\ncur. bps", 1 );
			XcpsmonColumnHeader( "cps_send\navg. bps", 1 );
			XcpsmonColumnHeader( "#gets", 1 );
			XcpsmonColumnHeader( "cps_get\ncur. bps", 1 );
			XcpsmonColumnHeader( "cps_get\navg. bps", 1 );
			XcpsmonColumnHeader( NULL, 1 );
			XcpsmonProcessLine( "< 2K", 1 );
			XcpsmonProcessLine( "< 8K", 1 );
			XcpsmonProcessLine( "< 32K", 1 );
			XcpsmonProcessLine( ">= 32K", 1 );
			XcpsmonFooter( "Running..." );
		}
	}

	mon_refresh();
}

/******************************************************************/
/*
	FUNCTION NAME:

	 main

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

main(argc,argv)
int argc;
char **argv;
{
	char fname[80],*c;
	int prognum = 0, version = 0;
	/* int no_job_name = FALSE; */
	int opt;
	FILE *fp, *file;
	int ch;
	int newx,newy;
	int i,j;
	char buf[80];
	int size;
	int addr;

	suppress_errors = -1;

	if( xcps_mon )
	{
		strcpy( monitor_name, "xcps_mon" );

		if( ( argc == 2 && !strcmp( argv[1], "-help" ) ) || ( argc == 1 ) )
		{
			printf(XUSAGE);
			exit(0);
		}

		xcpsmon = TRUE;
		XcpsmonInit( &argc, argv );
		XcpsmonCommandLine( job_hostname, &prognum, &mon_type, &mon_proc,
			&mon_class, &rpcvers );
		if( argc > 2 )
		{
			printf(XUSAGE);
			exit(0);
		}
		/*
		if( fork() )
		{
			fclose( stdin );
			fclose( stdout );
			fclose( stderr );
			exit(0);
		}
		*/
		if( prognum < MIN_PROGNUM )
			prognum = prognum + MIN_PROGNUM;
		version = 1;
		XcpsmonWindow();
	}
	else
	{
		strcpy( monitor_name, "cps_mon" );

/* Check the number of arguments */

		if( ( argc == 2 && !strcmp( argv[1], "-help" ) ) || ( argc == 1 ) )
		{
			printf(USAGE);
			exit(0);
		}

/* Get the options */

		while ((opt = getopt(argc,argv,"dufsqtp:c:a:h:i:")) != -1)
		{
			switch ((char) opt)
			{
				case 'd':
					mon_type = MON_DELTA_CALLS;
					break;
				case 'u':
					mon_type = MON_USER_STATUS;
					break;
				case 'f':
					mon_type = MON_CALLS;		
					break;
				case 's':
					mon_type = MON_STATES;
					break;
				case 'q':
					mon_type = MON_QUEUES;
					break;
				case 't':
					mon_type = MON_TRATES;
					break;
				case 'p':
					mon_proc = atoi(optarg);
					break;
				case 'c':
					mon_class = atoi(optarg);
					break;
				case 'a':
				case 'h':
					rpcvers = TRUE;
					/* no_job_name = TRUE; */
					sprintf(job_hostname,"%s",optarg);
					if( strncmp( argv[optind], "-", 1 ) )
					{
						prognum = atoi(argv[optind++]);
						if( prognum < MIN_PROGNUM )
							prognum = prognum + MIN_PROGNUM;
					}
					version = 1;
					break;
				case 'i':
					prognum = atoi(optarg);
					if( prognum < MIN_PROGNUM )
						prognum = prognum + MIN_PROGNUM;
					break;
				case '?':
					printf(USAGE);
					exit(0);
			}
		}
	}

/* Read the mdf file */

/* if( !no_job_name )
		sprintf(job_name,"%s",argv[optind]); */

/* determnine mdf file name for SHM version */
	if( !rpcvers )
	{
		sprintf(fname,"%s.mdf",argv[optind]);
		sprintf(job_name,"%s",argv[optind]);
	}

/* Connect to RPC server */
	if( rpcvers )
	{
		create_rpc_client( job_name, job_hostname, &prognum, &version, &cl );
		if( cl == NULL )
		{
			printf("%s: failed to connect to RPC server\n",monitor_name);
			exit(0);
		}
		/* test for user's rights to this CPS job */
		check_for_access_rights = TRUE;
		rpc_get_jm_version_1( NULL, cl );
		check_for_access_rights = FALSE;
	}
	else
		gethostname( job_hostname, 64 );

#ifndef TEST

	if( !rpcvers )
	{
		fp = fopen(fname,"r");
		if (fp == NULL)
		{
			printf("%s: failed to open %s\n",monitor_name,fname);
			exit(0);
		}
		fscanf(fp,"%d",&shmid);
		fclose(fp);
	}

#endif

/* Attach to the shared memory */

#ifdef TEST

	mon_shm = (struct cps_job_stats *) hmalloc(sizeof(struct cps_job_stats));
	bzero(mon_shm,sizeof(struct cps_job_stats));
	mon_shm->num_proc = 12;
	for (i=0; i<4; i++) mon_shm->class[i] = 1;
	for (i=4; i<8; i++) mon_shm->class[i] = 2;
	for (i=8; i<12; i++) mon_shm->class[i] = 3;

#else

	if( rpcvers )
	{
		mon_shm =
			(struct cps_job_stats *) hmalloc(sizeof(struct cps_job_stats));
		bzero(mon_shm,sizeof(struct cps_job_stats));
	}
	else
	{
		addr = (int) shm_attach(shmid);
		if (addr == -1)
		{
			printf("%s: failed to attach to shared memory\n",monitor_name);
			exit(0);
		}
		mon_shm = (struct cps_job_stats *) addr;
	}

#endif

/* Set up a control-C handler */

	sig_establish_handler(SIGINT,handle_ctrlc);

/* Copy the shared memory to the local copy */

	mon_loc = *mon_shm;

	if( rpcvers )
	{
		mon_loc.num_proc = *rpc_get_num_proc_1( NULL, cl );
		for( i = 0; i < mon_loc.num_proc; i++ )
			mon_loc.class[i] = *rpc_get_process_class_1( &i, cl );
	}

/* Initialize the display */

	suppress_errors = 0;
	mon_init();
	init_done = -1;
	mon_refresh();
	force_updates();

/* Check for changes every second */

	if( !xcpsmon )
	{
		sig_establish_handler(SIGALRM,handle_alarm);
		alarm(5);
	}
	else
	{
		XcpsmonTimeout();
	}

/* Monitor the job */

	if( xcpsmon )
		XcpsmonEventLoop();

	while (1)
	{
		ch = getc( stdin );  /* getc(stdin) is used because of its 'BLOCKING' 
					ability which reduces cpu usage when there is no
					user interaction */
		if (ch == 'q') break;
		if (ch == 'r')
		{
			sig_block_one(SIGALRM);
			mon_refresh();
			sig_pop_mask();
			continue;
		}
		newy = pad_y;
		if ((ch != 'j') && (ch != 'k')) continue;
		if (ch == 'j')
			newy = pad_y + 1;
		if (ch == 'k')
			newy = pad_y - 1;
		newy = (newy < 0) ? 0 : newy;
		newy = (newy > (pad_rows-24)) ? (pad_rows-24) : newy;
		if (newy != pad_y)
		{
			pad_y = newy;
			sig_block_one(SIGALRM);
		    mon_refresh();
			sig_pop_mask();
		}
	}

/* Exit from the monitor */

	mon_exit();
	
}
