/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: hash.c

    Explanation: Simple hash table lookup for storing 32-bit integers.

    Exported Routines:


*/

#include <stdio.h>

#define TABLE_SIZE 256
#define BUCKET_SIZE 16

static struct table_entry {

	int nitems;				/* Number of items stored here */
	int data[BUCKET_SIZE];
	int arg[BUCKET_SIZE];
	int valid[BUCKET_SIZE];		/* -1 = valid, 0 = invalid */
	struct table_entry *next;		/* Overflow */

};

static struct table_entry htable[TABLE_SIZE];
static int init_flag = -1;

static int hashfunc(x)
int x;
{
	x /= 4;
	return x & (TABLE_SIZE-1);
}

static int htbl_init()
{
	int i;

	for (i=0; i<TABLE_SIZE; i++) {
		htable[i].nitems = 0;
		htable[i].next = NULL;
	}
	init_flag = 0;
}

/************************************************************************/
/*

    Function: htbl_lookup

    Description: Find a hash table entry.

    Function Value: NULL pointer if not found, otherwise pointer to
		    table_entry record.

*/
/************************************************************************/

struct table_entry *htbl_lookup(x,n)
int x,*n;
{
	int i;
	struct table_entry *h;

	if (init_flag) htbl_init();
	i = hashfunc(x);
	h = &(htable[i]);
	while (h != NULL) {
		for (i=0; i<(h->nitems); i++) {
			if (h->valid[i] && (x == h->data[i])) {
				if (n != NULL) *n = h->arg[i];
				return h;
			}
		}
		h = h->next;
	}
	return h;

}

/************************************************************************/
/*

    Function: htbl_makevalid

    Description: Change an invalid entry to valid.

    Function Value: -1 if change took place, 0 if not.

*/
/************************************************************************/

int htbl_makevalid(x)
int x;
{
	int i;
	struct table_entry *h;

	if (init_flag) htbl_init();
	i = hashfunc(x);
	h = &(htable[i]);
	while (h != NULL) {
		for (i=0; i<(h->nitems); i++) {
			if (h->valid[i] == 0) {
				if (h->data[i] == x) {
					h->valid[i] = -1;
					return -1;
				}
			}
		}
		h = h->next;
	}
	return 0;
}

/************************************************************************/
/*

    Function: htbl_dump

    Description: Print contents of the hash table

    Arguments: none

    Function Value: none

*/
/************************************************************************/

int htbl_dump()
{
	int i,j,k;
	struct table_entry *h;

	printf("\nHash Table:\n");
	if (init_flag) htbl_init();
	for (i=0; i<TABLE_SIZE; i++) {
		h = &(htable[i]);
		if (h->nitems > 0) printf("%d: ",i);
		k = 0;
		while (h != NULL) {
			for (j=0; j<(h->nitems); j++) {
				k++;
				printf("%d:%d ",h->data[j],h->valid[j]);
			}
			h = h->next;
		}
		if (k > 0) printf("\n");
	}
	fflush(stdout);
}

/************************************************************************/
/*

    Function: htbl_insert

    Description: Inserts an integer into the hash table.

    Function Value: -1 if entry already exists, otherwise 0 for success.

*/
/************************************************************************/

int htbl_insert(x,n)
int x,n;
{
	int i;
	struct table_entry *h,*newh;

	if (init_flag) htbl_init();
	if (htbl_lookup(x,NULL) != NULL) {
		return -1;		/* Already there */
	}
	if (htbl_makevalid(x)) {
		return 0;
	}
	i = hashfunc(x);
	h = &(htable[i]);
	while (h->nitems == BUCKET_SIZE) h = h->next;
	i = h->nitems;
	h->data[i] = x;
	h->arg[i] = n;
	h->valid[i] = -1;
	h->nitems++;
	if (h->nitems == BUCKET_SIZE) {
		newh = (struct table_entry *) malloc(sizeof(struct table_entry));
		newh->next = NULL;
		newh->nitems = 0;
		h->next = newh;
	}
	return 0;

}

/************************************************************************/
/*

    Function: htbl_remove

    Description: Marks a hash table entry as invalid.

    Arguments:

    Function Value: -1 if entry is missing, 0 for success.

*/
/************************************************************************/

int htbl_remove(x)
int x;
{
	int i;
	struct table_entry *h,*newh;

	if (init_flag) htbl_init();
	if ((h = htbl_lookup(x,NULL)) == NULL) return -1;	/* Entry is missing */
	for (i=0; i<(h->nitems); i++) {
		if (x == h->data[i]) {
			h->valid[i] = 0;
			return 0;
		}
	}
	abort();
}
