/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.	    
	   All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or	     
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

	Source File: hmalloc.c

	Explanation: Records addresses returned by malloc() and flags an bad
		 addresses given to free().

		 Sanity check for negative numbers passed to malloc.

		 Options available to keep track of total amount of memory
		 allocated, and for periodic reporting.

	Exported Routines: hmalloc, hfree


*/

#include <stdio.h>
#include <signal.h>
#include "../inc/perr.h"

#define FACILITY "HMALLOC"

int hm_nmalloc = 0;		/* Number of calls to hmalloc */
int hm_nfree = 0;		/* Number of calls to hfree */
int hm_mem = 0;			/* Amount of memory allocated */

int hm_report = 0;		/* -1 turns on periodic reporting */
int hm_op_interval = 100;	/* Report every 100 calls */
int hm_mem_interval = 10000000;	/* Report every 10 megabytes */

int hm_op_count = 0;
int hm_mem_count = 0;

char *malloc();


/************************************************************************/
/*

	Function: hmalloc

	Description: Calls malloc and records allocated address in a hash table.

*/
/************************************************************************/

char *hmalloc(n)
int n;
{
	char *c;
	char msg[80];

	if (n < 0) {
		printerr(FACILITY,FATAL_MSG,"Negative parameter passed to malloc",
		    NULL,2);
		abort();
	}
	else if (n == 0) {
		return( 0 );
	}

	sig_block_all();
	c = malloc(n);

	if (c != NULL) {
		htbl_insert(c,n);
	}
	else {
		printerr(FACILITY,FATAL_MSG,
		    "Malloc failed - no available memory or memory corrupted",
		    NULL,10);
		exit(0);
	}
	sig_pop_mask();

	/* Do periodic reporting if necessary */

	hm_nmalloc++;
	hm_mem += n;
	if (hm_report) {
		hm_op_count++;
		if (hm_op_count >= hm_op_interval) {
			sprintf(msg,"Number of calls to malloc = %d, free = %d",
			    hm_nmalloc,hm_nfree);
			printerr(FACILITY,INFO_MSG,msg,NULL,3);
			sprintf(msg,"Total memory allocated = %d bytes",hm_mem);
			printerr(FACILITY,INFO_MSG,msg,NULL,4);
			hm_op_count -= hm_op_interval;
		}
		hm_mem_count += n;
		if (hm_mem_count >= hm_mem_interval) {
			sprintf(msg,"Number of calls to malloc = %d, free = %d",
			    hm_nmalloc,hm_nfree);
			printerr(FACILITY,INFO_MSG,msg,NULL,8);
			sprintf(msg,"Total memory allocated = %d bytes",hm_mem);
			printerr(FACILITY,INFO_MSG,msg,NULL,9);
			hm_mem_count -= hm_mem_interval;
		}
	}

	return c;
}

/************************************************************************/
/*

	Function: hfree

	Description: Calls free, but first checks to see if the address is in
		 the hash table.

*/
/************************************************************************/

int hfree(addr)
char *addr;
{
	int n;

	if( addr == NULL ) {
		return( 0 );
	}

	sig_block_all();

	if (htbl_lookup(addr,&n) == NULL) {
		printerr(FACILITY,FATAL_MSG,"Attempted to free non-allocated address",
		    NULL,1);
		abort();
	}
	else {
		htbl_remove(addr);	/* Remove the entry from the table */
		free(addr);
	}

	hm_nfree++;
	hm_mem -= n;

	sig_pop_mask();
	return 0;
}

/************************************************************************/
/*

	Function: hm_dump

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int hm_dump() {
	char msg[80];
	sprintf(msg,"Total amount of memory allocated: %d\n",hm_mem);
	printerr(FACILITY,INFO_MSG,msg,NULL,5);
	sprintf(msg,"Total number of calls to malloc:  %d\n",hm_nmalloc);
	printerr(FACILITY,INFO_MSG,msg,NULL,6);
	sprintf(msg,"Total number of calls to free:    %d\n",hm_nfree);
	printerr(FACILITY,INFO_MSG,msg,NULL,7);
}
