/* machup.c

 Author:        Bill Meyer & Matt Fausey
                Fermi National Accelerator Laboratory - Mail Station 369
                P.O Box 500
                Batavia, IL 60510
                Internet: meyer@fnfaus.fnal.gov
			  fausey@fnfaus.fnal.gov

 Copyright (c) 1992 Universities Research Association, Inc.
       All Rights Reserved

    *** Must be root to execute this subroutine.

    This subroutine sends a very low-level ICMP (Internet Control Message 
Protocol) ECHO REQUEST to a specified host and waits for a reply.  ICMP 
messages are encapsulated in IP datagrams.  The ICMP ECHO REQUEST message 
format looks like this:

   +-----------+-------------+---------------------------------------------+
   | IP header | ICMP header | ICMP data                                   |
   +-----------+-------------+---------------------------------------------+

    The semantics of a raw socket are as follows.  We create a raw Internet
socket and specify that we want to use the ICMP protocol.  The KERNEL PREPENDS
an appropriate IP header to any data that we write to this socket.  The
protocol field in this IP header is the value that we specified when the
socket was created.  Whenever data is received by the kernel for the
ICMP protocol, a copy of the data is passed to ALL PROCESSES that have created
raw sockets for the ICMP protocol.  This data contains the IP header that was
received.  Our receiving routine must check two things- that the received data
is an ICMP ECHO REPLY, and that it was sent by this process.

***************************************************************************/

#include <stdio.h>
#include <errno.h>
#include "../inc/time.h"
#include "../inc/types.h"
#include "../inc/socket.h"
#include <sys/signal.h>
#include <sys/ioctl.h>

#define SIZE_ICMP_HDR		8
#define SIZE_TIME_DATA		8
#define DEF_DATALEN		56
#define PACKETSIZE		SIZE_ICMP_HDR+DEF_DATALEN
#define MAXPACKET		4096

static int in_chksum(ptr,nbytes)
unsigned short *ptr;
int nbytes;
{
	long sum;
	unsigned short oddbyte;

	sum = 0;
	while(nbytes > 1) {
		sum += *ptr++;
		nbytes -= 2;
	}

	if(nbytes == 1) {
		oddbyte = 0;
		*((unsigned char *)&oddbyte) = *(unsigned char *)ptr;
		sum += oddbyte;
	}

	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += (sum >> 16);

	return(~sum);
}

int machup(host)
char *host;
{
	unsigned char sendpacket[MAXPACKET], recvpacket[MAXPACKET];
	int sock;
	struct hostent *hp;
	struct protoent *proto;
	struct sockaddr_in remote;
	struct icmp *icp;
	struct ip *ip;
	struct timeval start, stop, poll_t;
	int i, n, pid, recvlen, triptime, ipheaderlen;
	int timeout = 0;

	fd_set fds,fds1;

	poll_t.tv_sec = 3;		/* Wait 3 seconds for data to arrive */
	poll_t.tv_usec = 0;

	if(!(hp = gethostbyname(host))) return 0;

	remote.sin_family = hp->h_addrtype;
	bcopy(hp->h_addr, (caddr_t)&remote.sin_addr, hp->h_length);

	pid = getpid() & 0xFFFF;

	if(!(proto = getprotobyname("icmp"))) return 0;

	if((sock = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) return 0;

	FD_ZERO(&fds);
	FD_SET(sock,&fds);

	/* Create the ICMP header */

	icp = (struct icmp *)sendpacket;
	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_id = pid;
	icp->icmp_seq = 0;

	/* Record the time at which we sent the packet */

	gettimeofday(&start, (struct timezone *)0);
	bcopy(&start, &sendpacket[SIZE_ICMP_HDR], sizeof(start));

	/* Set each byte of ICMP data to an arbitrary number */

	for(i = SIZE_TIME_DATA; i < DEF_DATALEN; i++)
		sendpacket[SIZE_ICMP_HDR+i] = i;

	/* Now compute and store the checksum for the ICMP packet */

	icp->icmp_cksum = in_chksum( (unsigned short *)icp, PACKETSIZE);

	/* Send the packet */

	n = sendto(sock, sendpacket, PACKETSIZE, 0, (struct sockaddr *)&remote,
	    sizeof(remote));

	/* Check the status of the send */

	if(n < 0 || n != PACKETSIZE) goto machdown;

	/* Wait for a reply */

poll_sock:
	fds1 = fds;
	n = select(sock+1,&fds1,(fd_set *) 0, (fd_set *) 0,&poll_t);
	if (n < 0) {
		if (errno == EINTR) goto poll_sock;
		goto machdown;
	}
	if (n == 0) goto machdown;		/* timeout */

	recvlen = sizeof(remote);
	if((n = recvfrom(sock, recvpacket, sizeof(recvpacket), 0,
	    (struct sockaddr *)&remote, &recvlen)) < 0) goto machdown;

	gettimeofday(&stop, (struct timezone *)0);
	ip = (struct ip *)recvpacket;
	ipheaderlen = ip->ip_hl << 2;
	if(n < ipheaderlen + ICMP_MINLEN) goto machdown;

	n -= ipheaderlen;

	icp = (struct icmp *)(recvpacket + ipheaderlen);

	if(icp->icmp_seq == 1) goto machdown;

	close(sock);
	return -1;

machdown:
	close(sock);
	return 0;
}
/*

void check_machine(host)
char *host;
{
    if (machup(host)) {
	printf("%s is up\n",host);
    }
    else {
	printf("%s is down\n",host);
    }
}

main()
{

    char host[132];

    while (1) {
    	printf("Enter hostname: ");
    	scanf("%s",host);
    	printf("\n");
    	check_machine(host);
	fflush(stdout);
    }

}

*/
