/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: pmcomm.c

    Explanation: Routines used by clients to communicate with the PM server.

    Exported Routines:  pm_connect
			pm_disconnect
			send_to_pm
			read_from_pm
			set_batch_job
			add_spec_request
			add_vsn_request
			add_list_request
			send_mount_requests
			get_spec_reply
			get_vsn_reply
			get_list_reply
			handle_db_update

    Global Variables:   retry_flag	- Control automatic reconnect to PM
			db		- CPS Batch database
			rcv_tmounts	- Subscribe to tape mounts,
					  tape replies, and oper msgs.
			pm_sock		- Socket for communicating with PM.
			asynch_flag	- Allow read_from_pm to return after
					  reading only a db_update.
			connected_to_pm - Non-zero if connected to a PM.

    Notes:

	All communications regarding obtaining the database and processing
	updates are contained entirely within this package.

*/

#include <stdio.h>
#include "../inc/perr.h"
#include "../inc/db_param.h"
#include "../inc/db_struct.h"
#include "../inc/pm_param.h"
#include "../inc/pm.h"
#include "../inc/msgqueue.h"

#ifdef VMS
#include descrip
#define SETUP_DSC(descr,string) descr.dsc$w_length = strlen(string); \
                                descr.dsc$a_pointer = string; \
                                descr.dsc$b_class = DSC$K_CLASS_S; \
                                descr.dsc$b_dtype = DSC$K_DTYPE_T;
#endif

#define FACILITY "PMCOMM"
#define SLEEPTIME 5


int pm_sock = -1;		/* Socket for communicating with the PM */
int pm_port = PM_PORT;		/* Port number PM is listening at */
int connected_to_pm = 0;	/* True if currently connected */
int once_connected = 0;		/* True if we've ever been connected to a PM */
char pm_hostname[64];		/* Hostname where PM is located */
char pm_domain[80];		/* YP Domain for locating PM */
int reconnect_flag;		/* True if we've lost the connection */


struct database db;
char *msg_expected = NULL;	/* Message we are expecting next */
int retry_flag = -1;		/* Retry flag which can be set by client */
int rcv_tmounts = 0;		/* Non-zero will cause mount requests and
				   replies to be forwarded to this client */

int asynch_flag = 0;		/* Return from read_from_pm even if only
				   a db_update was read */
int db_clear_flag = -1;		/* True if we should clear the database
				   on connecting to a PM */
int get_batch_domain = -1;	/* If true, get the batch domain from 
				   the environment variable BATCH_DOMAIN */
int multi_pm_interface = 0;	/* True if multi_pm interface is being used */

static struct msg_queue *deferred = NULL;
static struct msg_queue *tmp_q = NULL;

static int db_initialized = 0;
static int bj_flag = 0;		/* Non-zero if this client associated with
				   a batch job */

static struct batchjob bj;	/* Batch job with which we are associated */
static struct mountreq *qmrq = NULL;	/* Mount requests queued but not sent */
static struct mountreq *qmrq_tail = NULL;
static struct mountreq *pmrq = NULL;	/* Mount requests pending */
static struct mountreq *pmrq_tail = NULL;

static char errmsg[256];

extern char *msg_cmd,**msg_args;
extern int suppress_errors;

struct mountreq *find_mount_request();


/************************************************************************/
/*

    Function:  set_batch_job

*/
/************************************************************************/

int set_batch_job(psname,entry,pipestage)
char *psname;
int entry,pipestage;
{
	bj_flag = -1;
	strcpy(bj.prodsys,psname);
	bj.entry = entry;
	bj.pipestage = pipestage;
	return 0;
}

/************************************************************************/
/*

    Private Functions:
			collect_db_updates
			get_database

*/
/************************************************************************/

static int collect_db_updates()
{
	int e;
	char buf[256];

	while (1) {
		if (read_counted(pm_sock,buf,256) != 0) {
			if (!retry_flag) {
				printerr(FACILITY,ERROR_MSG,
				    "Failed to collect database updates",NULL,1);
			}
			return -1;
		}
		msg_parse(buf);

		if (multi_pm_interface) {
			m_pm_startup_msg();
		}

		if (strcmp(msg_cmd,"db_update") != 0) {
			sprintf(errmsg,"Unexpected message: %s",
			    msg_build(msg_cmd,msg_args));
			printerr(FACILITY,ERROR_MSG,errmsg,NULL,2);
			return -1;
		}

		/* Ignore error messages */

		set_suppress(-1);
		if (database_update(&db.fs,&db.cpu,&db.tape,&db.ps,msg_args) == -1){
			restore_suppress();
			return 0;
		}
		restore_suppress();
	}
}

static int get_database()
{

	/* Request the job information */

	if (send_counted(pm_sock,build_query_msg(DB_QUERY_DB,NULL,NULL)) != 0) {
		if (!retry_flag) {
			printerr(FACILITY,ERROR_MSG,"Failed to request database",NULL,3);
		}
		return -1;
	}

	/* Read database updates until all information has arrived */

	if (collect_db_updates() == -1) return -1;

	sort_tape_drives(&(db.tape));

	return 0;
}

/************************************************************************/
/*

    Function:  check_pm_connection

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int check_pm_connection()
{
	if (!connected_to_pm) {
		reconnect_flag = -1;
		return pm_connect();
	}
	else {
		return 0;
	}
}

/************************************************************************/
/*

    Function:  pm_connect

    Description:  Connects to a PM server.

    Function Value:  0 = success, -1 = failure.

*/
/************************************************************************/

int pm_connect()
{

	struct mountreq *m;

	char bj_msg[132];

	int status;

	int i;

#ifdef VMS
	char logname[80],logval[80];
	int defined;
	struct dsc$descriptor_s logname_d,logval_d;
#endif

	/* Initialize the database if it hasn't been */

	if (!db_initialized) {
		db.cpu = NULL;
		db.fs = NULL;
		db.tape = NULL;
		db.ps = NULL;
		db_initialized = -1;
	}

	/* Create a message queue for deferred messages */

	if (deferred == NULL) {
		deferred = create_queue();
		tmp_q = create_queue();
	}

	/* Find the YP-domain for looking up the hostname of the PM */

#ifndef VMS

	if (get_batch_domain && get_yp_domain(pm_domain) == -1) {
		printerr(FACILITY,ERROR_MSG,"Failed to get YP domain",NULL,5);
		return(-1);
	}

#endif

start:

	/* Create an active socket */

	if (pm_sock != -1) socket_close(pm_sock);

	pm_sock = create_active_socket();
	if (pm_sock == -1) {
		printerr(FACILITY,ERROR_MSG,"Failed to create active socket",
		    NULL,4);
		return(-1);
	}

	/* Erase the entire database before re-connecting */

	if (db_clear_flag) clear_database(&db.fs,&db.cpu,&db.tape,&db.ps);

	/* Find the location of the PM */

#ifndef VMS

	if (map_lookup(PMLOC_FILE,pm_domain,
	    "locations","pm_host",pm_hostname,64) == -1) {
		printerr(FACILITY,ERROR_MSG,"Failed to get PM location",NULL,6);
		return(-1);
	}

#else

	/* Location of the PM is set by a logical name */

	strcpy(logname,"PM_HOST");
	SETUP_DSC(logname_d,logname);
	for (i=0; i<79; i++) logval[i] = ' ';
	logval[79] = '\0';
	SETUP_DSC(logval_d,logval);
	str_translate_logical(&logname_d,&defined,&logval_d);
	if (!defined) {
		printerr(FACILITY,FATAL_MSG,"PM_HOST logical not defined",NULL,7);
		exit(0);
	}
	strncpy(logval,logval_d.dsc$a_pointer,logval_d.dsc$w_length);
	logval[logval_d.dsc$w_length] = '\0';
	logval[79] = '\0';
	for (i=0; i<79; i++) {
		if ((logval[i] == ' ') ||  (logval[i] == '\t')) {
			logval[i] = '\0';
			break;
		}
	}
	str_tolower(logval);
	strcpy(pm_hostname,logval);
#endif

	/* Check to see if the machine is up */

	if (geteuid() == 0) {
		if (!machup(pm_hostname)) {
			sprintf(errmsg,"%s not responding",pm_hostname);
			printerr(FACILITY,ERROR_MSG,errmsg,NULL,79);
			return -1;
		}
	}

	/* Attempt to connect to the PM */

	status = initiate_connection(pm_sock,pm_hostname,pm_port);
	if (status == -1) {
		if (retry_flag && once_connected) {
			sleep_1(SLEEPTIME);
			goto start;
		}
		else {
			socket_close(pm_sock);
			pm_sock = -1;
			sprintf(errmsg,"Failed to connect to PM on %s",pm_hostname);
			printerr(FACILITY,ERROR_MSG,errmsg,NULL,30);
			return -1;
		}
	}

	/* Tell the PM what job we are with */

	if (bj_flag) {
		sprintf(bj_msg,"my_job|%s|%d|%d",bj.prodsys,bj.entry,bj.pipestage);
	}
	else {
		sprintf(bj_msg,"my_job|none");
	}
	status = send_counted(pm_sock,bj_msg);
	if (status != 0) {
		if (retry_flag && once_connected) {
			sleep_1(SLEEPTIME);
			goto start;
		}
		else {
			socket_close(pm_sock);
			pm_sock = -1;
			printerr(FACILITY,ERROR_MSG,"Failed to connect to PM",NULL,8);
			return -1;
		}
	}

	/* Get the entire database from the PM */

	if (get_database() == -1) {
		if (retry_flag && once_connected) {
			sleep_1(SLEEPTIME);
			goto start;
		}
		else {
			socket_close(pm_sock);
			pm_sock = -1;
			printerr(FACILITY,ERROR_MSG,"Failed to connect to PM",NULL,9);
			return -1;
		}
	}

	/* Subscribe to mount requests and replies if necessary */

	if (rcv_tmounts) {
		if (send_counted(pm_sock,"rcv_tmounts") != 0) {
			if (retry_flag && once_connected) {
				sleep_1(SLEEPTIME);
				goto start;
			}
			else {
				socket_close(pm_sock);
				pm_sock = -1;
				printerr(FACILITY,ERROR_MSG,"Failed to connect to PM",NULL,10);
				return -1;
			}
		}
	}

	/* Tell PM what mount requests we are waiting for */

	m = pmrq;
	while (m != NULL) {
		if (send_counted(pm_sock,build_mount_waiting(m)) != 0) {
			if (retry_flag && once_connected) {
				sleep_1(SLEEPTIME);
				goto start;
			}
			else {
				socket_close(pm_sock);
				pm_sock = -1;
				printerr(FACILITY,ERROR_MSG,"Failed to connect to PM",NULL,13);
				return -1;
			}
		}
		m = m->next;
	}

	once_connected = -1;
	connected_to_pm = -1;
	return(0);

}

/************************************************************************/
/*

    Function:  pm_disconnect

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int pm_disconnect() {
	socket_close(pm_sock);
	pm_sock = -1;
	connected_to_pm = 0;
	return 0;
}

/************************************************************************/
/*

    Function:  send_to_pm

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int send_to_pm(msg)
char *msg;
{
start:
	if (check_pm_connection() == -1) {
		printerr(FACILITY,ERROR_MSG,
		    "Message not sent, lost connection with PM",NULL,11);
		return -1;
	}
	if (send_counted(pm_sock,msg) != 0) {
		if (retry_flag && once_connected) {
			pm_disconnect();
			sleep_1(SLEEPTIME);
			goto start;
		}
		else {
			pm_disconnect();
			return -1;
		}
	}
	return 0;
}

/************************************************************************/
/*

    Function:  handle_db_update

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int handle_db_update()
{
	int e;

	/* If it is a database update, update the database and get another message */

	if (strcmp(msg_cmd,"db_update") == 0) {
		set_suppress(-1);
		database_update(&db.fs,&db.cpu,&db.tape,&db.ps,msg_args);
		restore_suppress();
		return -1;
	}
	return 0;
}

/************************************************************************/
/*

    Function:  read_from_pm

    Description:

    Arguments:

    Function Value:

	 0:	Non-db_update message received.
	 1: 	db_update message received.
	 2:	No messages, reconnected to another PM.
	-1:	Lost connection, failed to reconnect.

*/
/************************************************************************/

int read_from_pm() {

	char buf[256];
	struct queue_element *e,*ematch,*alloc_queue_elem();
	int qflag;

start:
	reconnect_flag = 0;
	if (check_pm_connection() == -1) {
		printerr(FACILITY,ERROR_MSG,
		    "No messages read, lost connection with PM",NULL,12);
		return -1;
	}
	if (reconnect_flag) return 2;

	/* Check to see if we have some deferred messages */

	if (deferred->front != NULL) {

		/* If we are not expecting any specific type of message, pull off the first
   message and process it.  Otherwise, pull of the first message that matches
   the message expected. */

		if (msg_expected == NULL) {
			e = remove_q(deferred);
			strcpy(buf,e->msg);
			dealloc_queue_elem(e);
			goto lookatit;
		}
		else {
			ematch = NULL;
			while (deferred->front != NULL) {
				e = remove_q(deferred);
				qflag = -1;
				if (ematch == NULL) {
					msg_parse(e->msg);
					if (strcmp(msg_cmd,msg_expected) == 0) {
						qflag = 0;
						ematch = e;
					}
				}
				if (qflag) enter_q(tmp_q,e);
			}
			while (tmp_q->front != NULL) {
				e = remove_q(tmp_q);
				enter_q(deferred,e);
			}
			if (ematch != NULL) {
				strcpy(buf,ematch->msg);
				dealloc_queue_elem(ematch);
				goto lookatit;
			}
		}
	}

	/* Get a message from the PM */

	if (read_counted(pm_sock,buf,256) != 0) {
		if (retry_flag && once_connected) {
			pm_disconnect();
			sleep_1(SLEEPTIME);
			goto start;
		}
		else {
			pm_disconnect();
			return -1;
		}
	}

	/* Look at the message */

lookatit:
	msg_parse(buf);
	if (handle_db_update() == -1) {
		if (asynch_flag) return 1;
		goto start;
	}
	if (msg_expected == NULL) return 0;

	/* If we are expecting a specific message, defer all messages that
   don't match */

	if (strcmp(msg_cmd,msg_expected) != 0) {
		e = alloc_queue_elem();
		strcpy(e->msg,buf);
		enter_q(deferred,e);
		goto start;
	}

	return 0;
}

/************************************************************************/
/*

    Function:  add_spec_request
	       add_vsn_request
	       add_list_request
	       add_user_request

*/
/************************************************************************/

int add_spec_request(cpuname,vsn,prot,tdrive,naux,aux)
char *vsn,*tdrive,*cpuname;
int prot,naux,*aux;
{
	struct mountreq *m;
	int i;

	m = (struct mountreq *) hmalloc(sizeof(struct mountreq));
	bzero(m,sizeof(struct mountreq));
	m->type = DB_SPEC_MOUNT;
	strcpy(m->vsn,vsn);
	strcpy(m->dev,tdrive);
	strcpy(m->cpu,cpuname);
	strcpy(m->user,"x");
	m->rstate = prot;
	for (i=0; i<naux; i++) m->aux[i] = aux[i];
	add_to_mreq_list(&qmrq,&qmrq_tail,m);
	return 0;
}

int add_vsn_request(cpuname,dtype,vsn,prot,naux,aux)
char *vsn,*cpuname;
int dtype,prot,naux,*aux;
{
	struct mountreq *m;
	int i;

	m = (struct mountreq *) hmalloc(sizeof(struct mountreq));
	bzero(m,sizeof(struct mountreq));
	m->type = DB_VSN_MOUNT;
	strcpy(m->vsn,vsn);
	strcpy(m->user,"x");
	m->rstate = prot;
	m->devtype = dtype;
	strcpy(m->cpu,cpuname);
	for (i=0; i<naux; i++) m->aux[i] = aux[i];
	add_to_mreq_list(&qmrq,&qmrq_tail,m);
	return 0;
}

int add_list_request(cpuname,class,listnum,naux,aux)
char *cpuname;
int class,listnum,naux,*aux;
{
	struct mountreq *m;
	int i;

	m = (struct mountreq *) hmalloc(sizeof(struct mountreq));
	bzero(m,sizeof(struct mountreq));
	m->type = DB_LIST_MOUNT;
	m->class = class;
	m->listnum = listnum;
	strcpy(m->cpu,cpuname);
	strcpy(m->user,"x");
	for (i=0; i<naux; i++) m->aux[i] = aux[i];
	add_to_mreq_list(&qmrq,&qmrq_tail,m);
	return 0;
}

int add_user_request(username,cpuname,dtype,vsn,prot,naux,aux)
char *vsn,*cpuname,*username;
int dtype,prot,naux,*aux;
{
	struct mountreq *m;
	int i;

	m = (struct mountreq *) hmalloc(sizeof(struct mountreq));
	bzero(m,sizeof(struct mountreq));
	m->type = DB_USER_MOUNT;
	strcpy(m->vsn,vsn);
	strcpy(m->user,username);
	m->rstate = prot;
	m->devtype = dtype;
	strcpy(m->cpu,cpuname);
	for (i=0; i<naux; i++) m->aux[i] = aux[i];
	add_to_mreq_list(&qmrq,&qmrq_tail,m);
	return 0;
}

/************************************************************************/
/*

    Function: send_mount_requests

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int send_mount_requests()
{
	struct mountreq *m;
	int rstatus;
	int a;

	/* Send all mount requests queued */

	while (qmrq != NULL) {

		/* Get a mount request id */

send_idreq:
		if (send_to_pm("m_idreq") != 0) return -1;
		a = asynch_flag;
		asynch_flag = 0;
		msg_expected = "mount_id";
		rstatus = read_from_pm();
		msg_expected = NULL;
		asynch_flag = a;
		if (rstatus == 2) goto send_idreq;
		if (rstatus == -1) return -1;
		if (strcmp(msg_cmd,"mount_id") != 0) {
			sprintf(errmsg,"Expected mount id, got: %s",
			    msg_build(msg_cmd,msg_args));
			printerr(FACILITY,ERROR_MSG,errmsg,NULL,20);
			return -1;
		}
		qmrq->id = atoi(msg_args[0]);

		/* Send the mount request */

		if (send_to_pm(build_mount_request(qmrq)) != 0) return -1;
		m = qmrq;
		remove_from_mreq_list(&qmrq,&qmrq_tail,m);
		add_to_mreq_list(&pmrq,&pmrq_tail,m);
	}

	return 0;
}

/************************************************************************/
/*

    Function: get_spec_reply
	      get_vsn_reply
	      get_list_reply
	      get_user_reply

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int get_spec_reply(cpuname,vsn,prot,tdrive,rstatus,errmsg)
char *cpuname,*vsn,*tdrive,*errmsg;
int *prot,*rstatus;
{
	struct mountreq m,*m1;
	int rdstatus;
	int a;

	/* Get a message (should be a specific mount reply) */

	a = asynch_flag;
	asynch_flag = 0;
	while ((rdstatus = read_from_pm()) == 2);	/* Re-read on reconnects */
	asynch_flag = a;
	if (rdstatus != 0) return rdstatus;

	/* Make sure we got a mount reply */

	if (strcmp(msg_cmd,"m_reply") != 0) {
		sprintf(errmsg,"Expected mount reply, got: %s",
		    msg_build(msg_cmd,msg_args));
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,14);
		return -1;
	}

	/* Decode the reply */

	decode_mount_reply(&m);

	/* Make sure it is a specific mount reply */

	if (m.type != DB_SPEC_MOUNT) {
		printerr(FACILITY,ERROR_MSG,"Did not get correct mount reply",NULL,15);
		return -1;
	}

	/* Copy reply to the arguments */

	strcpy(vsn,m.vsn);
	strcpy(tdrive,m.dev);
	strcpy(errmsg,m.errmsg);
	*prot = m.rstate;
	*rstatus = m.rstatus;
	strcpy(cpuname,m.cpu);

	/* Remove from the list of pending mounts */

	m1 = find_mount_request(pmrq,m.id);
	if (m1 == NULL) {
		printerr(FACILITY,ERROR_MSG,"Stray mount reply",NULL,26);
		return -1;
	}
	remove_from_mreq_list(&pmrq,&pmrq_tail,m1);
	hfree(m1);

	return 0;
}

int get_vsn_reply(cpuname,vsn,prot,dtype,tdrive,rstatus,errmsg,naux,aux)
char *cpuname,*vsn,*tdrive,*errmsg;
int *prot,*rstatus,*dtype,naux,*aux;
{
	struct mountreq m,*m1;
	int i,rdstatus;
	int a;

	/* Get a message (should be a vsn mount reply) */

	a = asynch_flag;
	asynch_flag = 0;
	while ((rdstatus = read_from_pm()) == 2);	/* Re-read on reconnects */
	asynch_flag = a;
	if (rdstatus != 0) return rdstatus;

	/* Make sure we got a mount reply */

	if (strcmp(msg_cmd,"m_reply") != 0) {
		sprintf(errmsg,"Expected mount reply, got: %s",
		    msg_build(msg_cmd,msg_args));
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,16);
		return -1;
	}

	/* Decode the reply */

	decode_mount_reply(&m);

	/* Make sure it is a vsn mount reply */

	if (m.type != DB_VSN_MOUNT) {
		printerr(FACILITY,ERROR_MSG,"Did not get correct mount reply",NULL,17);
		return -1;
	}

	/* Copy reply to the arguments */

	strcpy(vsn,m.vsn);
	strcpy(tdrive,m.dev);
	strcpy(errmsg,m.errmsg);
	strcpy(cpuname,m.cpu);
	*prot = m.rstate;
	*rstatus = m.rstatus;
	*dtype = m.devtype;
	for (i=0; i<naux; i++) aux[i] = m.aux[i];

	/* Remove from the list of pending mounts */

	m1 = find_mount_request(pmrq,m.id);
	if (m1 == NULL) {
		printerr(FACILITY,ERROR_MSG,"Stray mount reply",NULL,22);
		return -1;
	}
	remove_from_mreq_list(&pmrq,&pmrq_tail,m1);
	hfree(m1);

	return 0;
}

int get_list_reply(cpuname,vsn,prot,tdrive,rstatus,errmsg,
class,listnum,listpos,naux,aux)
char *cpuname,*vsn,*tdrive,*errmsg;
int *prot,*rstatus,*class,*listnum,*listpos,naux,*aux;
{
	struct mountreq m,*m1;
	int i,rdstatus;
	int a;

	/* Get a message (should be a list mount reply) */

	a = asynch_flag;
	asynch_flag = 0;
	while ((rdstatus = read_from_pm()) == 2);	/* Re-read on reconnects */
	asynch_flag = a;
	if (rdstatus != 0) return rdstatus;

	/* Make sure we got a mount reply */

	if (strcmp(msg_cmd,"m_reply") != 0) {
		sprintf(errmsg,"Expected mount reply, got: %s",
		    msg_build(msg_cmd,msg_args));
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,18);
		return -1;
	}

	/* Decode the reply */

	decode_mount_reply(&m);

	/* Make sure it is a list mount reply */

	if (m.type != DB_LIST_MOUNT) {
		printerr(FACILITY,ERROR_MSG,"Did not get correct mount reply",NULL,19);
		return -1;
	}

	/* Copy reply to the arguments */

	strcpy(vsn,m.vsn);
	strcpy(tdrive,m.dev);
	strcpy(errmsg,m.errmsg);
	strcpy(cpuname,m.cpu);
	*prot = m.rstate;
	*rstatus = m.rstatus;
	for (i=0; i<naux; i++) aux[i] = m.aux[i];
	*class = m.class;
	*listnum = m.listnum;
	*listpos = m.listpos;

	/* Remove from the list of pending mounts */

	m1 = find_mount_request(pmrq,m.id);
	if (m1 == NULL) {
		printerr(FACILITY,ERROR_MSG,"Stray mount reply",NULL,21);
		return -1;
	}
	remove_from_mreq_list(&pmrq,&pmrq_tail,m1);
	hfree(m1);

	return 0;
}

int get_user_reply(user,cpuname,vsn,prot,dtype,tdrive,rstatus,errmsg,naux,aux)
char *cpuname,*vsn,*tdrive,*errmsg,*user;
int *prot,*rstatus,*dtype,naux,*aux;
{
	struct mountreq m,*m1;
	int i,rdstatus;
	int a;

	/* Get a message (should be a user mount reply) */

	a = asynch_flag;
	asynch_flag = 0;
	while ((rdstatus = read_from_pm()) == 2);	/* Re-read on reconnects */
	asynch_flag = a;
	if (rdstatus != 0) return rdstatus;

	/* Make sure we got a mount reply */

	if (strcmp(msg_cmd,"m_reply") != 0) {
		sprintf(errmsg,"Expected mount reply, got: %s",
		    msg_build(msg_cmd,msg_args));
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,16);
		return -1;
	}

	/* Decode the reply */

	decode_mount_reply(&m);

	/* Make sure it is a user mount reply */

	if (m.type != DB_USER_MOUNT) {
		printerr(FACILITY,ERROR_MSG,"Did not get correct mount reply",NULL,27);
		return -1;
	}

	/* Copy reply to the arguments */

	strcpy(vsn,m.vsn);
	strcpy(tdrive,m.dev);
	strcpy(errmsg,m.errmsg);
	strcpy(cpuname,m.cpu);
	strcpy(user,m.user);
	*prot = m.rstate;
	*rstatus = m.rstatus;
	*dtype = m.devtype;
	for (i=0; i<naux; i++) aux[i] = m.aux[i];

	/* Remove from the list of pending mounts */

	m1 = find_mount_request(pmrq,m.id);
	if (m1 == NULL) {
		printerr(FACILITY,ERROR_MSG,"Stray mount reply",NULL,28);
		return -1;
	}
	remove_from_mreq_list(&pmrq,&pmrq_tail,m1);
	hfree(m1);

	return 0;
}

/************************************************************************/
/*

    Function:  remove_mount_request

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int remove_mount_request(m)
struct mountreq *m;
{
	struct mountreq *m1;

	/* Remove from the list of pending mounts */

	m1 = find_mount_request(pmrq,m->id);
	if (m1 == NULL) {
		printerr(FACILITY,ERROR_MSG,"Stray mount reply",NULL,25);
		return -1;
	}
	remove_from_mreq_list(&pmrq,&pmrq_tail,m1);
	hfree(m1);
}
