/*************************************************************************
**************************************************************************
**************************************************************************
******                                                                ****
******  MXYZPTLK:  A C++ implementation of differential algebra.      ****
******                                                                ****
******  Copyright (c) 1990  Universities Research Association, Inc.   ****
******                All Rights Reserved                             ****
******                                                                ****
******  Author:    Leo Michelotti                                     ****
******                                                                ****
******             Fermilab                                           ****
******             P.O.Box 500                                        ****
******             Mail Stop 345                                      ****
******             Batavia, IL   60510                                ****
******                                                                ****
******             Phone: (708) 840 4956                              ****
******             Email: michelot@hazel.fnal.gov                     ****
******                    michelotti@adcalc.fnal.gov                  ****
******                    almond::michelotti                          ****
******                                                                ****
******  Release    Version 1.0 : January   31, 1990                   ****
******  Dates:              .1 : October   11, 1990                   ****
******                      .2 : November   6, 1990                   ****
******                      .3 : October    2, 1992                   ****
******                      .4 : July      20, 1993                   ****
******                     2.0   August    12, 1993                   ****
******                     3.0   May       19, 1994                   ****
******                      .1   September 14, 1994                   ****
******                                                                ****
**************************************************************************
**************************************************************************
******                                                                ****
******  This material resulted from work developed under a            ****
******  Government Contract and is subject to the following           ****
******  license: The Government retains a paid-up, nonexclusive,      ****
******  irrevocable worldwide license to reproduce, prepare           ****
******  derivative works, perform publicly and display publicly       ****
******  by or for the Government, including the right to distribute   ****
******  to other Government contractors.  Neither the United          ****
******  States nor the United States Department of Energy, nor        ****
******  any of their employees, makes any warranty, express or        ****
******  implied, or assumes any legal liability or responsibility     ****
******  for the accuracy, completeness or usefulness of any           ****
******  information, apparatus, product, or process disclosed, or     ****
******  represents that its use would not infringe privately owned    ****
******  rights.                                                       ****
******                                                                ****
******  These files are made avaliable for use by specific            ****
******  individuals or a specific group.  They are not to be          ****
******  reproduced or redistributed.                                  ****
******                                                                ****
******  These files are provided "as is" with no warranties of any    ****
******  kind including the warranties of design,                      ****
******  merchantibility and fitness for a particular purpose,         ****
******  or arising from a course of dealing, usage or trade           ****
******  practice.                                                     ****
******                                                                ****
******  These files are provided with no support or obligation on     ****
******  the part of U.R.A. or Fermilab to assist in their use,        ****
******  correction, modification or enhancement.                      ****
******                                                                ****
******  Neither U.R.A. nor Fermilab shall be liable with              ****
******  respect to the infringement of copyrights, trade              ****
******  secrets or any patents by these files or any part             ****
******  thereof.                                                      ****
******                                                                ****
******  In no event will U.R.A., Fermilab, or the author(s) be        ****
******  liable for any lost revenue or profits or other               ****
******  special, indirect and consequential damages, even if          ****
******  they have been advised of the possibility of such             ****
******  damages.                                                      ****
******                                                                ****
**************************************************************************
*************************************************************************/

#ifndef JL_HXX
#define JL_HXX

typedef char (*MX_R_FUNCPTR)(const int*, const double);

// *******************************************************************
struct JLtermData;
struct JLterm {    

  // Data
  int* index;	  // An integer array giving the derivatives associated
                  //      with the JLterm.  For example, ( 1, 1, 0, 2 )
                  //            would correspond to D_1^1 D_2^1 D_4^2 .
  int weight;     // The sum of the values in index.  For the above example,
                  //                                        this would be 4.
  double value;   // The value associated with the JLterm.

  // Constructors and destructors
  JLterm(); 
  JLterm( int*, double );
  JLterm( JLterm* );
  JLterm( JLterm& );
  JLterm( JLtermData& );
  ~JLterm();

  // Operators
  void operator=( const JLterm& );
  double coeff(){ return value; }
} ;


// *******************************************************************
struct JL : public dlist {  

  int count;        // The number of JL terms in the variable
  int weight;       // The maximum weight of the terms
  int accuWgt;      // Highest weight computed accurately
  double* refPoint; // Point about which derivatives are to be expanded.
  int rc;           // Reference counting for garbage collection

  // Private member functions_________________________________________
  void insert( JLterm* );
  void append( JLterm* );
  void peekAt( int );
  void printCoeffs( int );
  JLterm* remove( dlink* );

  // Constructors and destructors_____________________________________
  JL();
  JL( const double );
  JL( const JL& );
  JL( JL* );
  ~JL();

  // Public member functions__________________________________________
  JLterm* get();            // Pops the top term, which should be the 
                            // one of lowest weight.
  void addTerm( JLterm* );  // Public only for diagnostic purposes.
  void fixReference( const double* );
  void fixReference();
  void fixReference( JL& );
  void fixReferenceAtEnd( const LieOperator& );
  void fixReferenceAtStart( const LieOperator& );

  char isNilpotent();
  void peekAt();
  void printCoeffs();
  dlist image();
  void writeToFile( char*   /* Name of unopened file */ );
  void writeToFile( FILE* );

  void getReference( double* );
  void scaleBy( double );
  void setVariable( double, int );
  void setVariable( int );
  double standardPart();
  void clear();
  double weightedDerivative( int* );
  double derivative( int* );
  double operator()( double* );	   // Performs a multinomial evaluation of 
				   // the JL variable.  Essentially acts as a 
				   // power series expansion.
  JL& operator()( JL* );           // Self explanatory ...
  JL& operator()( LieOperator& );  // Self explanatory ...
  JL& D( int* )const ;		   // Performs differentiation of a JL variable.

  JL& operator=( const JL& );
  JL& operator=( const double& );
  JL& operator+=( const double& );

  JL& inverse();             // Returns the inverse jet
                             //  only when the number of
                             //  coordinates = 1.

} ;

struct JLtermData{
  int  numvar;
  int* index;
  int weight;
  double value;
  JLtermData();
  JLtermData( JLterm& );
  ~JLtermData();
};

void operator+=( JL&, const JL& );   // ??? Why are these void operators?
void operator-=( JL&, const JL& );
void operator*=( JL&, const JL& );
void operator/=( JL&, const JL& );
char operator!=( const JL&, const JL& );
char operator!=( const JL&, const double& );
char operator!=( const double&, const JL& );


// *******************************************************************
//
//     Setup routines, etc.
//

extern void JLFixReference( double* );   // ??? Still exists???
extern void JLFixReference( JL& );
extern void JLFixReferenceAtStart( const LieOperator& );
extern void JLFixReferenceAtEnd( LieOperator& );

extern ostream& operator<<(ostream&, JL&);
extern istream& operator>>(istream&, JL&);
extern char operator==( const JLterm&, const JLterm& );
extern char operator==( const JL&,     const JL& );
extern char operator==( const JL&,     const double& );
extern char operator==( const double&, const JL& );
extern char operator<=( JLterm&, JLterm& );
extern char operator%=( JLterm&, JLterm& );   // Acts like == but compares
                                       // indices only.
extern JLterm operator*( JLterm&, JLterm& );

#endif // JL_HXX
