/*************************************************************************
**************************************************************************
**************************************************************************
******                                                                ****
******  MXYZPTLK:  A C++ implementation of differential algebra.      ****
******                                                                ****
******  Copyright (c) 1990  Universities Research Association, Inc.   ****
******                All Rights Reserved                             ****
******                                                                ****
******  Author:    Leo Michelotti                                     ****
******                                                                ****
******             Fermilab                                           ****
******             P.O.Box 500                                        ****
******             Mail Stop 345                                      ****
******             Batavia, IL   60510                                ****
******                                                                ****
******             Phone: (708) 840 4956                              ****
******             Email: michelot@hazel.fnal.gov                     ****
******                    michelotti@adcalc.fnal.gov                  ****
******                    almond::michelotti                          ****
******                                                                ****
******  Release    Version 1.0 : January   31, 1990                   ****
******  Dates:              .1 : October   11, 1990                   ****
******                      .2 : November   6, 1990                   ****
******                      .3 : October    2, 1992                   ****
******                      .4 : July      20, 1993                   ****
******                     2.0   August    12, 1993                   ****
******                     3.0   May       19, 1994                   ****
******                      .1   September 14, 1994                   ****
******                                                                ****
**************************************************************************
**************************************************************************
******                                                                ****
******  This material resulted from work developed under a            ****
******  Government Contract and is subject to the following           ****
******  license: The Government retains a paid-up, nonexclusive,      ****
******  irrevocable worldwide license to reproduce, prepare           ****
******  derivative works, perform publicly and display publicly       ****
******  by or for the Government, including the right to distribute   ****
******  to other Government contractors.  Neither the United          ****
******  States nor the United States Department of Energy, nor        ****
******  any of their employees, makes any warranty, express or        ****
******  implied, or assumes any legal liability or responsibility     ****
******  for the accuracy, completeness or usefulness of any           ****
******  information, apparatus, product, or process disclosed, or     ****
******  represents that its use would not infringe privately owned    ****
******  rights.                                                       ****
******                                                                ****
**************************************************************************
**************************************************************************
******                                                                ****
******  Notes:                                                        ****
******                                                                ****
******  11/6/90   .. Leo Michelotti                                   ****
******  JL functions  pow( JL&, double )  and  sqrt( JL& ) have       ****
******  been included in the lexicon.                                 ****
******                                                                ****
******                                                                ****
******  10/11/90  .. Leo Michelotti                                   ****
******  JLSetup routines have been changed so that the variable       ****
******  JL__oneEnv.JL__allSet and the array JL__oneEnv.JL__refPoint   ****
******  are set earlier.  This fixes a bug which first arose in       ****
******  an application involving concatenation.                       ****
******                                                                ****
******                                                                ****
******  10/2/92  .. Leo Michelotti                                    ****
******  In   double JL::weightedDerivative( int* ind ), the cast      ****
******  appearing in declaring a dlist_iterator was changed           ****
******  from (dlist) to (dlist&) in order to bypass a call to         ****
******  the dlist copy constructor.  I suspect g++ does not           ****
******  handle casting correctly, but I cannot yet prove it.          ****
******  It may also be true that the copy constructor is not          ****
******  working properly.                                             ****
******                                                                ****
******                                                                ****
******  7/20/93 .. Leo Michelotti                                     ****
******  In an attempt to improve efficiency, functions that           ****
******  previously returned JL have been changed to return            ****
******  JL&.  In conjunction with this, a disposable base             ****
******  class has been defined in order to provide for some           ****
******  primitive garbate collection.                                 ****
******                                                                ****
******  The code sandwiched between #ifdef POOLED and #endif          ****
******  POOLED statements is still under development.  It             ****
******  should not be activated by the Makefile.                      ****
******                                                                ****
******                                                                ****
******  8/12/93 .. Leo Michelotti                                     ****
******  The LieOperator methods .expMap() and .operator^              ****
******  have now been implemented.  The former                        ****
******  integrates equations of motion by developing                  ****
******  exponential maps, and the latter simply                       ****
******  evaluates a Lie derivative.  The one restriction              ****
******  at the moment is the assumption made by                       ****
******  .expMap() that the point of evaluation is a                   ****
******  fixed point of the vector field.  There are ways              ****
******  of getting around this -- but, another day.                   ****
******  Demos are contained in files Lie_X_Test.cc,                   ****
******  where X = A ... E.                                            ****
******                                                                ****
******     __ This upgrade is significant enough to                   ****
******  warrant a version change, from 1.X to 2.0.                    ****
******  Huzzah!                                                       ****
******                                                                ****
******     __ Addition has been modified so that small                ****
******  terms indistinguishable from zero are not                     ****
******  carried.  Further, arithmetic on LieOperators                 ****
******  has been enabled.                                             ****
******                                                                ****
**************************************************************************
*************************************************************************/

#include "mxyzptlk.hxx"

// ================================================================
//      External routines
//

extern char nexcom( int, int, int* );  
                                // Computes the next composition
                                //  of an integer into a number of parts.
                                //  Algorithm devised by Herbert Wilf.

extern "C" {
 int bcfRec( int, int );        // Recursive evaluation of binomial
                                //  coefficient.
 int nearestInteger( double );  // Returns the integer nearest to its
                                //  double argument.
}


// ================================================================

#ifndef MX_SMALL
#define MX_SMALL    1.0e-12 // Used by JL::addTerm to decide 
                            //   removal of a JLterm.
#endif  // MX_SMALL

#ifndef MX_MAXITER
#define MX_MAXITER  100     // Maximum number of iterations allowed
                            //   in iterative routines
#endif  // MX_MAXITER

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLFixReference( double* r ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) Jet::oneEnv.refPoint[i] = r[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLFixReferenceAtEnd( LieOperator& x ) {
 int i;
 double* u = new double [ Jet::oneEnv.NumVar ];
 x.standardPart( u );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) Jet::oneEnv.refPoint[i] = u[i];
 delete [] u;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLFixReferenceAtStart( const LieOperator& x ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) Jet::oneEnv.refPoint[i] = x.refPoint[i];
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Implementation of Class JL

//    Constructors and destructors    |||||||||||||||||||||||||||

JL::JL() {
 int i;
 
 rc = 1;
 weight = -1;
 count = 0;
 accuWgt = Jet::oneEnv.MaxWeight;
 
 if( Jet::oneEnv.allSet == 137 ) {
   refPoint = new double[ Jet::oneEnv.NumVar ];
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = Jet::oneEnv.refPoint[i];
 }
 else {
   refPoint = NULL;
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JL::JL( const double x ) {
 int* ndx;
 int i;
 
 if( Jet::oneEnv.allSet != 137 ) {
   printf( "\n*** JL::JL( double ) ERROR *** " );
   printf( "Prior call to JLSetup required.\n" );
   exit(0);
   }
 
 clear();
 rc = 1;
 weight = 0;
 count = 0;
 accuWgt = Jet::oneEnv.MaxWeight;
 
 refPoint = new double[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = Jet::oneEnv.refPoint[i];
 
                             // ??? There is some danger here!  This constructor
                             //     governs the conversion of double into JL.
                             //     In an expression like 3.0 * z, where z is a JL,
                             //     the multiplication may not go through if the
                             //     current reference point has changed since z
                             //     was created.  One way around this is to overload
                             //     all arithmetic operators explicitly so conversion
                             //     is never necessary.
 
 ndx = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) ndx[i] = 0;
 insert( new JLterm( ndx, x ) );
 delete [] ndx;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JL::JL( const JL& x ) {
 dlist_iterator getNext( (dlist&) x );
 JLterm* p;
 JLterm* q;
 int i;
 
 static char firstCall = 1;
 if( firstCall ) {
  firstCall = 0;
  cout << "\n*** WARNING *** The JL copy constructor was called for \n"
       <<   "*** WARNING *** some inexplicable reason.\n"
       <<   "*** WARNING *** Go figure!!\n\n";
  cout.flush();
 }

 if( Jet::oneEnv.allSet != 137 ) {
   printf( "\n*** JL::JL( JL& ) ERROR *** " );
   printf( "Prior call to JLSetup required.\n" );
   exit(0);
   }
 
 if( Jet::oneEnv.allSet != 137 ) {
   printf( "\n*** JL::JL( JL& ) ERROR *** " );
   printf( "Prior call to JLSetup required.\n" );
   exit(0);
   }
 
 while( p = (JLterm*) getNext() ) {
   q = new JLterm( p );
   append( q );
   }

 rc       = 1;
 count    = x.count;
 weight   = x.weight;
 accuWgt  = x.accuWgt;
 refPoint = new double[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JL::JL( JL* x ) {
 dlist_iterator getNext( *(dlist*) x );
 JLterm* p;
 JLterm* q;
 int i;
 
 static char firstCall = 1;
 if( firstCall ) {
  firstCall = 0;
  cout << "\n*** WARNING *** The JL pointer constructor was called for \n"
       <<   "*** WARNING *** some inexplicable reason.\n"
       <<   "*** WARNING *** Go figure!!\n\n";
  cout.flush();
 }

 if( Jet::oneEnv.allSet != 137 ) {
   printf( "\n*** JL::JL( JL& ) ERROR *** " );
   printf( "Prior call to JLSetup required.\n" );
   exit(0);
   }
 
 if( Jet::oneEnv.allSet != 137 ) {
   printf( "\n*** JL::JL( JL& ) ERROR *** " );
   printf( "Prior call to JLSetup required.\n" );
   exit(0);
   }
 
 while( p = (JLterm*) getNext() ) {
   q = new JLterm( p );
   append( q );
   }

 rc       = 1;
 count    = x->count;
 weight   = x->weight;
 accuWgt  = x->accuWgt;
 refPoint = new double[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = x->refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JL::~JL() {
 clear();
 delete [] refPoint;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

//     Member functions(public)  |||||||||||||||||||||||||||||||

void JL::addTerm( JLterm* a) {
 dlist_traversor getNext( *(dlist*) this );
 dlink* p;
 
 // If the value of *a is 0, don't bother with it unless
 // the weight is also 0.
 if( ( a -> value == 0.0 ) && ( a -> weight != 0 ) ) {
   delete a;
   return;
 }
 
 //
 // In the event that the candidate is lighter than the first
 // element in the JL dlist ... or the list is empty
 //
 p = getNext();
 
 if( p == 0 ) {
   insert( a );
   return;
   }
 
 if(     *a   <=   ( *(JLterm*) (p->info()) )   ) {
   if(   *a   %=   ( *(JLterm*) (p->info()) )   ) {
 
     (( (JLterm*) (p->info()) ) -> value ) += ( a -> value );
 
     if(
         (
           fabs( ( (JLterm*) (p->info()) ) -> value ) < MX_SMALL*fabs( a -> value )
         )
         // REMOVE ??? &&
         // REMOVE ??? (
         // REMOVE ???   p != ( dlist::lastPtr() ) -> nextPtr()
         // REMOVE ??? )
       )
 
         delete remove( p );
 
     delete a;
     return;
     }
 
   else {
     insert( a );
     return;
     }
 
   }
 
 //
 // Otherwise, search through the list
 //
 while( p = getNext() ){
 
   if(     *a   <=   ( *(JLterm*) (p->info()) )   ) {
     if(   *a   %=   ( *(JLterm*) (p->info()) )   ) {
 
       (( (JLterm*) (p->info()) ) -> value ) += ( a -> value );
 
       if(
	   (
	     fabs( ( (JLterm*) (p->info()) ) -> value ) < MX_SMALL*fabs( a -> value )
	   )
	   // REMOVE ??? &&
	   // REMOVE ??? (
	   // REMOVE ???   p != ( dlist::lastPtr() ) -> nextPtr()
	   // REMOVE ??? )
	 )
 
         delete remove( p );
 
       delete a;
       return;
       }
 
     else {
       p -> putAbove( new dlink( a, 0, 0 ) );
       (this->count)++;
       if( (this->weight) < (a->weight) ) this->weight = a->weight;
       return;
       }
 
     }
   }
 
 //
 //  The new entry is heavier than existing ones.
 //
 append( a );
 return;
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::fixReference( const double* r ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = r[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::fixReference() {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = Jet::oneEnv.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::fixReference( JL& x ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::fixReferenceAtEnd( const LieOperator& x ) {
 int i;
 double* u = new double [ Jet::oneEnv.NumVar ];
 x.standardPart( u );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = u[i];
 delete [] u;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::fixReferenceAtStart( const LieOperator& x ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::getReference( double* r ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) r[i] = refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char JL::isNilpotent() {
 dlist_iterator getNext( *(dlist*) this );
 JLterm* p = (JLterm*) getNext();
 if( p->weight == 0 && p->value != 0.0 ) return 0;
 else                                    return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::peekAt() {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLterm* p;
 dlink* q;
 printf( "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 printf( "Reference point:  \n" );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) printf( "%e  ", refPoint[i] );
 printf( "\n" );
 while( q = getNext() ) {
   p = (JLterm*) q->info();
   printf( "Weight: %d   Value: %e  || ",
           p -> weight, p -> value );
   printf("Addresses: %d %d %d : %d\n", q->prevPtr(), q, q->nextPtr(), p );
   printf("Index:  ");
   for( i = 0; i < Jet::oneEnv.NumVar; i++ )
     printf( "%d  ", (p -> index)[i] );
   printf("\n\n");
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::printCoeffs() {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLterm* p;
 dlink* q;
 cout << "\nCount  = " << count 
      << ", Weight = " << weight 
      << ", Max accurate weight = " << accuWgt << endl;
 cout << "Reference point: " << endl;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  cout << ::form( "%e  ", refPoint[i] );
 cout << "\n" << endl;
 while( q = getNext() ) {
   p = (JLterm*) q->info();
   if( p->weight > accuWgt ) break;
   cout << "Index:  ";
   for( i = 0; i < Jet::oneEnv.NumVar; i++ )
     cout << ::form( "%3d", (p -> index)[i] );
   cout << ::form( "   Value: %14.6le  ", p -> value ) << endl;
   }
 cout << "\n" << endl;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::peekAt( int  ) {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLterm* p;
 dlink* q;
 fprintf( Jet::scratchFile,
          "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 fprintf( Jet::scratchFile, "Reference point:  \n" );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
   fprintf( Jet::scratchFile, "%e  ", refPoint[i] );
 fprintf( Jet::scratchFile, "\n" );
 while( q = getNext() ) {
   p = (JLterm*) q->info();
   fprintf( Jet::scratchFile, "Weight: %d   Value: %e  || ",
           p -> weight, p -> value );
   fprintf( Jet::scratchFile, "Addresses: %d %d %d : %d\n",
                  q->prevPtr(), q, q->nextPtr(), p );
   fprintf( Jet::scratchFile, "Index:  ");
   for( i = 0; i < Jet::oneEnv.NumVar; i++ )
     fprintf( Jet::scratchFile, "%d  ", (p -> index)[i] );
   fprintf( Jet::scratchFile, "\n\n");
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::printCoeffs( int  ) {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLterm* p;
 dlink* q;
 fprintf( Jet::scratchFile, "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 fprintf( Jet::scratchFile, "Reference point:  " );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) fprintf( Jet::scratchFile, "%e  ", refPoint[i] );
 fprintf( Jet::scratchFile, "\n\n" );
 while( q = getNext() ) {
   p = (JLterm*) q->info();
   if( p->weight > accuWgt ) break;
   fprintf( Jet::scratchFile,"Index:  ");
   for( i = 0; i < Jet::oneEnv.NumVar; i++ )
     fprintf( Jet::scratchFile, "%3d ", (p -> index)[i] );
   fprintf( Jet::scratchFile, "   Value: %14.6e  ", p -> value );
   fprintf( Jet::scratchFile,"\n");
   }
 fprintf( Jet::scratchFile, "\n\n" );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

dlist JL::image() {       // ??? Problem: nothing is recorded about the
                          // ??? reference point.

 dlist_iterator getNext( *(dlist*) this );
 JLterm* q;
 dlist ret;

 // ??? for( i = 0; i < Jet::oneEnv.NumVar; i++ ) fprintf( Jet::scratchFile, "%e  ", refPoint[i] );

 while( q = (JLterm*) getNext() )
   ret.append( new JLtermData( *q ) );

 return ret;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::writeToFile( char* fileName ) {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLterm* p;
 dlink* q;
 
 Jet::scratchFile = fopen( fileName, "w" );
 
 fprintf( Jet::scratchFile,
          "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 fprintf( Jet::scratchFile, "Reference point:  \n" );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
   fprintf( Jet::scratchFile, "%e  ", refPoint[i] );
 fprintf( Jet::scratchFile, "\n" );
 while( q = getNext() ) {
   p = (JLterm*) q->info();
   fprintf( Jet::scratchFile, "Weight: %d   Value: %e  || ",
           p -> weight, p -> value );
   fprintf( Jet::scratchFile, "Addresses: %d %d %d : %d\n",
                  q->prevPtr(), q, q->nextPtr(), p );
   fprintf( Jet::scratchFile, "Index:  ");
   for( i = 0; i < Jet::oneEnv.NumVar; i++ )
     fprintf( Jet::scratchFile, "%d  ", (p -> index)[i] );
   fprintf( Jet::scratchFile, "\n\n");
   }
 
 fclose( Jet::scratchFile );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::writeToFile( FILE* filePtr ) {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLterm* p;
 dlink* q;
 
 fprintf( filePtr,
          "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 fprintf( filePtr, "Reference point:  \n" );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
   fprintf( filePtr, "%e  ", refPoint[i] );
 fprintf( filePtr, "\n" );
 while( q = getNext() ) {
   p = (JLterm*) q->info();
   fprintf( filePtr, "Weight: %d   Value: %e  || ",
           p -> weight, p -> value );
   fprintf( filePtr, "Addresses: %d %d %d : %d\n",
                  q->prevPtr(), q, q->nextPtr(), p );
   fprintf( filePtr, "Index:  ");
   for( i = 0; i < Jet::oneEnv.NumVar; i++ )
     fprintf( filePtr, "%d  ", (p -> index)[i] );
   fprintf( filePtr, "\n\n");
   }
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::scaleBy( double y ) {
 JLterm* p;
 dlist_iterator getNext( *(dlist*) this );
 
 if ( this->count < 1 ) {
   cout << "\n\n*** WARNING::JL::scaleBy: function invoked \n"
        <<     "    by null JL variable with address " << (int) this 
        << endl;
   return;
   }
 
 while( p = (JLterm*) getNext() )   p->value *= y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::setVariable( double x, int j ) {
 int* ndx;
 int i;
 
 clear();
 if( ( j < 0 ) || ( j >= Jet::oneEnv.NumVar ) ) return;
 
 ndx = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) ndx[i] = 0;
 ndx[j] = 1;
 insert( new JLterm( ndx, 1.0 ) );
 delete [] ndx;
 
 ndx = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) ndx[i] = 0;
 insert( new JLterm( ndx, x ) );
 delete [] ndx;
 
 Jet::oneEnv.refPoint[j] = x;
 refPoint[j] = x;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::setVariable( int j ) {
 int* ndx;
 int i;
 double x;
 
 clear();
 if( ( j < 0 ) || ( j >= Jet::oneEnv.NumVar ) ) return;
 
 x = Jet::oneEnv.refPoint[j];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = Jet::oneEnv.refPoint[i];
 
 ndx = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) ndx[i] = 0;
 ndx[j] = 1;
 insert( new JLterm( ndx, 1.0 ) );
 delete [] ndx;
 
 ndx = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) ndx[i] = 0;
 insert( new JLterm( ndx, x ) );
 delete [] ndx;
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::insert( JLterm* a) {
 dlist::insert(a);
 count++;
 if( weight < (a->weight) ) weight = a->weight;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm* JL::get() {
 JLterm* p;
 if( count <= 0 ) return 0;
 p = ((JLterm*) dlist::get());
 if( --count == 0 ) weight = -1;
 return p;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::append( JLterm* a) {
 dlist::append(a);
 count++;
 if( weight < (a->weight) ) weight = a->weight;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm* JL::remove( dlink* w ) {
 JLterm* a;
 dlink* predecessor;
 dlink* successor;
 int predWeight;
 int succWeight;
 
 predecessor = w->prevPtr();
 successor   = w->nextPtr();
 
 a = (JLterm*) dlist::remove( w );
 count--;
 
 if( count == 0 ) {
   weight = -1;
   return a;
   }
 
 if ( weight > a->weight ) return a;
 
 predWeight = ( (JLterm*) ( predecessor->info() ) ) -> weight;
 succWeight = ( (JLterm*) ( successor->info() ) ) -> weight;
 weight = predWeight;
 if( succWeight > weight ) weight = succWeight;
 return a;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

double JL::standardPart() {
 dlist_iterator g( (dlist&) *this );
 JLterm* p; 
 if( count < 1 )       return 0.0;
 p = (JLterm*) g();
 if( p->weight  == 0 ) return p->value;
                       return 0.0;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JL::clear() {
 JLterm* p;
 while( p = get() )  
   delete p;
 dlist::clear();
 count   = 0;
 weight  = -1;
 accuWgt = Jet::oneEnv.MaxWeight;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

double JL::weightedDerivative( int* ind ) {
 int i;
 JLterm* p;
 char theOne;
 
 if( count < 1 ) return 0.0;
 
 dlist_iterator getNext( (dlist&) *this );

 while( p = (JLterm*) getNext() ) {
   theOne = 1;
   for ( i = 0; i < Jet::oneEnv.NumVar; i++ )
     theOne = theOne && ( (p->index)[i] == ind[i] );
   if( theOne ) return p->value;
 }
 
 return 0.0;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

double JL::derivative( int* ind ) {
 double n;
 double multiplier;
 double d;
 int i;
 
 d = weightedDerivative( ind );
 if ( d == 0.0 ) return 0.0;
 
 multiplier = 1.0;
 for ( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   n = ind[i];
   while(n)  multiplier *= n--;
   }
 
 return multiplier*d;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

double JL::operator()( double* x ) {
 JLterm* p;
 double v;
 int w;
 int i;
 double term;
 double* u;
 dlist_iterator getNext( *(dlist*) this );
 
 u = new double[ Jet::oneEnv.NumVar ];   // ??? Check other places in the source
                                 // code where this was not done correctly!!
 
 // Subtract off the reference point.
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) u[i] = x[i] - refPoint[i];
 
 // Evaluate and store monomials.
 
 // The zeroth one.
 Jet::oneEnv.monomial[0] = 1.0;
 
 // For all higher weights ...
 for( w = 1; w <= weight; w++ )
 
   // Get the next set of exponents of weight w.
   while( nexcom( w, Jet::oneEnv.NumVar, Jet::oneEnv.exponent ) ) {
 
     // Find the first non-zero exponent.
     i = 0;
     while( !( Jet::oneEnv.exponent[i++] ) ) ;
     i--;
 
     // The value of the monomial associated with this composition
     // is obtained by multiplying a factor into a previously
     // computed monomial.
     ( Jet::oneEnv.exponent[i] )--;
     Jet::monoCode();
     term = Jet::oneEnv.monomial[ Jet::monoRank() ];
     ( Jet::oneEnv.exponent[i] )++;
     Jet::monoCode();
     Jet::oneEnv.monomial[ Jet::monoRank() ] = term * u[i];
 
     } // End while loop.
 
 // Monomials have been now stored at this point.
 // Now traverse the JL variable and evaluate.
 
 v = 0.0;
 
 while( p = (JLterm*) getNext() ) {
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) Jet::oneEnv.exponent[i] = (p->index)[i];
   Jet::monoCode();
   v += ( p->value ) * ( Jet::oneEnv.monomial[ Jet::monoRank() ] );
   }
 
 delete [] u;
 
 return v;
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||



//***************************************************************
//***************************************************************
//***************************************************************
//
//      The overloaded operators for class JL
//
istream& operator>>(istream& is,  JL& x) {  // ??? This function shouldn't
  char buf[100];                            // ??? be here.
  int i,j;
  int count;
  double value;
  JLterm* q;
  int* index = new int[Jet::oneEnv.NumVar];

  x.clear();

  is >> buf;
  is >> buf;
  is >> count;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> x.weight;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> x.accuWgt;
  is >> buf;
  is >> buf;
  for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
   is >> x.refPoint[i];
  for( j = 0; j < count; j++) {
    is >> buf;
    for( i = 0; i < Jet::oneEnv.NumVar; i++ )  {
     is >> index[i];
    }
    is >> buf;
    is >> value;
    q = new JLterm(index,value);
    x.append( q );
  }

 delete [] index;
 return is;
}


ostream& operator<<(ostream& os, JL& x) {
 dlist_traversor getNext( (dlist&) x );
 int i;
 JLterm* p;
 dlink* q;

 os << "\n Count = " << x.count << " , Weight = " << x.weight;
 os << " , Max accurate weight = " << x.accuWgt << endl;
 os << "Reference point: ";
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
   os << setprecision(30) << x.refPoint[i] << " ";
 os << "\n\n";

 while( q = getNext() ) {
   p = (JLterm*) q->info();
   os << "Index: ";
   for( i = 0; i < Jet::oneEnv.NumVar; i++ )
     os << p -> index[i] << " ";
  os << "   Value: " << setprecision(30) << p -> value << endl;
 }
 return os << "\n" << endl;
}

char operator==( const JL& x, const JL& y ) {
 dlist_iterator getNextX( (dlist&) x );
 dlist_iterator getNextY( (dlist&) y );
 JLterm* p;
 JLterm* q;
 int i;
 
 if( ( x.count   != y.count )   ||
     ( x.weight  != y.weight )  ||
     ( x.accuWgt != y.accuWgt )
   ) return 0;
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
   if( x.refPoint[i] != y.refPoint[i] ) return 0;
 
 while( p = (JLterm*) getNextX() ) {
   if( !( q = (JLterm*) getNextY() ) ) return 0;
   if( !( *p == *q ) ) return 0;
   }
 
 if( (JLterm*) getNextY() ) {
   printf( "\n*** ERROR: Inconsistency in JL operator==\n" );
     // This point should never be reached, since that would imply that
     // x.count != y.count
   return 0;
   }
 
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const JL& x, const double& y ) {
 char result = 1;
 dlist_iterator getNext( (dlist&) x );
 JLterm* p;

 if( x.count < 1 ) {
  if( y == 0.0 ) return 1;
  else           return 0;
 }

 while( p = (JLterm*) getNext() ) 
  result = result && ( p->weight == 0 ? p->value == y : 
                                        p->value == 0.0 );

  // REMOVE ??? result = result && ( ( ( p->weight == 0 ) && ( p->value == y   ) )
  // REMOVE ???                      ||
  // REMOVE ???                      ( ( p->weight != 0 ) && ( p->value == 0.0 ) )
  // REMOVE ???                    )
  // REMOVE ??? 
 return result;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const double& y, const JL& x ){
 return x == y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const JL& x, const JL& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const JL& x, const double& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const double& x, const JL& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JL& JL::operator=( const double& x ) {
 int* ndx;
 int i;
 
 if( Jet::oneEnv.allSet != 137 ) {
   printf( "\n*** JL& JL::operator=( const double& x ) { ERROR *** " );
   printf( "Prior call to JLSetup required.\n" );
   exit(0);
   }
 
 clear();
 weight = 0;   // rc is purposely left untouched!!
 count = 0;
 accuWgt = Jet::oneEnv.MaxWeight;
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = Jet::oneEnv.refPoint[i];

                       // ??? There is some danger here!  This operator
                       //     governs the assignment of a double to JL.
 
 ndx = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) ndx[i] = 0;
 insert( new JLterm( ndx, x ) );
 delete [] ndx;
 return *this;
}
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JL& JL::operator=( const JL& x ) {
 dlist_iterator getNext( (dlist&) x );
 JLterm* p;
 JLterm* q;
 int i;
 
 if( this == &x ) return *this;
 
 clear();
 while( p = (JLterm*) getNext() ) {
   q = new JLterm( p );
   append( q );
   }

 count   = x.count;     // rc is purposely left untouched
 weight  = x.weight;
 accuWgt = x.accuWgt;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
 return *this;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JL& JL::operator+=( const double& x ) {   // ??? Untested!!
 static char firstNull = 1;
 int i;
 dlist_iterator getNext( *(dlist*) this );
 JLterm* p;                               
 JLterm* q;

 if( p = (JLterm*) getNext() )
 {
   if ( p->weight == 0 ) { 
    p->value += x;
    if( fabs( p->value ) <  MX_SMALL*fabs(x) ) p->value = 0.0;
   }
 
   else {
     q = new JLterm;
     for( i = 0; i < Jet::oneEnv.NumVar; i++ ) q->index[i] = 0;
     q->weight = 0;
     q->value = x;
     insert( q );
   }
 }
 else
 {
   if( firstNull ) {
     cerr << "*** WARNING ***                                 \n"
             "*** WARNING *** JL& JL::operator+=( const double& )\n"
             "*** WARNING *** Operator invoked by a null      \n"
             "*** WARNING *** JL.  This warning will not be   \n"
             "*** WARNING *** repeated.                       \n"
             "*** WARNING ***                                 \n"
          << endl;
     firstNull = 0;
   }
   operator=( x );
 }

 return *this;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm operator*( JLterm& x, JLterm& y ) {  
 JLterm z;
 int i;
 z.weight = x.weight + y.weight;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
   z.index[i] = x.index[i] + y.index[i];
 z.value = x.value * y.value;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Implementation of Class JLterm

JLterm::JLterm() {
 weight = 0;
 index = new int[ Jet::oneEnv.NumVar ];
 value = 0.0;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm::JLterm( int* l, double x ) {
 int i, dpt;
 
 dpt = 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   if( (l[i] < 0) ) {
     printf( "\n*** JL ERROR *** Bad index in JLTerm.\n" );
     exit(0);
     }
   dpt += l[i];
   }
 if( dpt > Jet::oneEnv.MaxWeight ) {
   printf("\n*** JL ERROR *** Attempt to load a JLTerm with too large a weight.\n");
   exit(0);
   }
 
 index = new int[ Jet::oneEnv.NumVar ];
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) index[i] = l[i];
 weight = dpt;
 value = x;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm::JLterm( JLterm* x ) {
 int i;
 
 weight = x->weight;
 value = x->value;
 index = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) index[i] = (x->index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm::JLterm( JLterm& x ) {
 int i;
 
 weight = x.weight;
 value = x.value;
 index = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) index[i] = (x.index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm::JLterm( JLtermData& x ) {
 int i;
 
 weight = x.weight;
 value = x.value;
 index = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) index[i] = (x.index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm::~JLterm() {
 delete [] index;
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Implementation of struct JLtermData

JLtermData::JLtermData() {
 printf( "\n" );
 printf( "*** ERROR ***                                   \n" );
 printf( "*** ERROR *** JLtermData::JLtermData()          \n" );
 printf( "*** ERROR *** Cannot declare a JLtermData       \n" );
 printf( "*** ERROR *** in this way.                      \n" );
 printf( "*** ERROR ***                                   \n" );
 exit(0);
}

JLtermData::JLtermData( JLterm& x ) {
 numvar = Jet::oneEnv.NumVar;   // ??? This has got to be changed!!
 index = new int[ numvar ];
 for( int i = 0; i < numvar; i++ ) index[i] = x.index[i];
 weight = x.weight;
 value = x.value;
}

JLtermData::~JLtermData() {
 if( index != 0 ) delete [] index;
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Overloaded operators for class JLterm

void JLterm::operator=( const JLterm& x ) {
 int i;
 weight = x.weight;
 value = x.value;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) index[i] = (x.index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const JLterm& a, const JLterm& b ) {
 int i;
 if( ( a.weight != b.weight) || ( a.value != b.value)  ) return 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
   if( ( a.index[i] != b.index[i]) ) return 0;
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const JLterm& a, const JLterm& b ) {
 return !( a == b );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator<=( JLterm& a, JLterm& b ) {
 int i;
 if( a.weight != b.weight ) return ( a.weight < b.weight );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   if( a.index[i] == b.index[i] ) continue;
   return ( a.index[i] < b.index[i] );
   }
 return 1;  // when all indices are the same.
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator%=( JLterm& a, JLterm& b ) {
 int i;
 if( a.weight != b.weight ) return 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
   if( ( a.index[i] != b.index[i]) ) return 0;
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
